// ****************************************************************
//
//   IBM Confidential
//   OCO Source Materials
//   <<PART NUMBER - 5724-B75>>
//   (C) Copyright IBM Corp. 2001,2002 - All Rights Reserved.
//   US Government Users Restricted Rights - Use, duplication or disclosure
//   restricted by GSA ADP Schedule Contract with IBM Corp.
//
// ****************************************************************

package com.ibm.etools.marshall.util;

/**
 * Utilities to aid in the marshalling of data
 */

import java.io.*;
import java.math.*;
import java.util.HashMap;

import com.ibm.etools.marshall.util.bidi.*;
import org.apache.wsif.format.WSIFFormatHandler;

public class ConversionUtils
{
   private static String copyright =
      "Licensed Material - Property of IBM " +
      "<<PART NUMBER - 5724-B75>> " +
      "(C) Copyright IBM Corp. 2001,2003 - All Rights Reserved. " +
      "US Government Users Restricted Rights - Use, duplication or disclosure " +
      "restricted by GSA ADP Schedule Contract with IBM Corp.";

  // ---------------- S390 Float routines ------------------------------------

  // IEEE 754 Floating point format represents double precision numbers as follows:
  // seee eeee | eeee ffff | ffff ffff | ffff ffff | ......
  // where s = sign bit
  // e = eponent bit
  // f = fraction bit
  // The decimal point is assumed at the left of the second byte.
  // When the exponent is non-zero, the MSB of the second byte is assumed to be
  // value 1 (2^-1 = 0.5).
  private final static long DOUBLE_SIGN_MASK         = 0x8000000000000000L;
  private final static long DOUBLE_EXPONENT_MASK     = 0x7ff0000000000000L;
  private final static long DOUBLE_MANTISSA_MASK     = 0x000fffffffffffffL;
  private final static long DOUBLE_MANTISSA_MSB_MASK = 0x0010000000000000L;

  // The exponent is treated as an unisgned number
  // A bias of 1022 is subtracted from it to give the power used.
  private final static long DOUBLE_BIAS = 1022;

  // The value of the IEEE Float is then given by mantissa * (2^(exp-1022)), the
  // sign bit giving the sign

  // S390 double precision floats have the following format:
  // seee eee | ffff ffff | ffff ffff | ffff ffff | ....
  // The exponent a power of 16, not 2.  Hence the best we can guarantee is that one
  // of the top four bits of the mantissa are non-zero - therefore there is no implied
  // bit as with IEEE
  // The exponent has a bias of 64, giving a range of -63..64
  private final static int  S390_DOUBLE_BIAS          = 64;
  private final static long S390_DOUBLE_EXPONENT_MASK = 0x7f00000000000000L;
  private final static long S390_DOUBLE_MANTISSA_MASK = 0x00ffffffffffffffL;

  // IEEE 754 Floating point format represents numbers as follows:
  // seee eeee | efff ffff | ffff ffff | ffff ffff
  // where s = sign bit
  // e = eponent bit
  // f = fraction bit
  // The decimal point is assumed at the left of the second byte.
  // When the exponent is non-zero, the MSB of the second byte is assumed to be
  // value 1 (2^-1 = 0.5).
  private final static int FLOAT_SIGN_MASK         = 0x80000000;
  private final static int FLOAT_EXPONENT_MASK     = 0x7f800000;
  private final static int FLOAT_MANTISSA_MASK     = 0x007fffff;
  private final static int FLOAT_MANTISSA_MSB_MASK = 0x00800000;

  // The exponent is treated as an unisgned number, giving a range of 0..255.
  // A bias of 126 is subtracted from it to give the power used.
  private final static int FLOAT_BIAS = 126;

  // The value of the IEEE Float is then given by mantissa * (2^(exp-126)), the
  // sign bit giving the sign

  // S390 floats have the following format:
  // seee eee | ffff ffff | ffff ffff | ffff ffff
  // The exponent a power of 16, not 2.  Hence the best we can guarantee is that one
  // of the top four bits of the mantissa are non-zero - therefore there is no implied
  // bit as with IEEE
  // The exponent has a bias of 64, giving a range of -63..64
  private final static int S390_FLOAT_BIAS          = 64;
  private final static int S390_FLOAT_EXPONENT_MASK = 0x7f000000;
  private final static int S390_FLOAT_MANTISSA_MASK       = 0x00ffffff;

// Packed Decimal Constants

private final static byte PACKED_DEC_POSOR = 0x0C;
private final static byte PACKED_DEC_NEGOR = 0x0D;
private final static byte PACKED_DEC_NOSIGN = 0x0F;

// External (Zoned) Decimal Constants

// EBCDIC
private final static byte EBCDIC_EXTDEC_POSOR = (byte)0xC0;
private final static byte EBCDIC_EXTDEC_NEGOR = (byte)0xD0;
private final static byte EBCDIC_ZERO_PAD = (byte)0xF0;
private final static byte EBCDIC_PLUS = 0x4E;
private final static byte EBCDIC_MINUS = 0x60;

// ASCII
private final static byte ASCII_EXTDEC_POSOR = 0x30;
private final static byte ASCII_EXTDEC_NEGOR = 0x70;
private final static byte ASCII_ZERO_PAD = 0x30;
private final static byte ASCII_PLUS = 0x2B;
private final static byte ASCII_MINUS = 0x2D;

// EBCDIC_CUSTOM
private final static byte EBCDIC_CUST_EXTDEC_POSOR = 0x40;
private final static byte EBCDIC_CUST_EXTDEC_NEGOR = 0x49;
private final static byte EBCDIC_CUST_ZERO_PAD = (byte)0x30;
private final static byte EBCDIC_CUST_PLUS = 0x2B;
private final static byte EBCDIC_CUST_MINUS = 0x2D;
private final static byte[] EBCDIC_CUST_POS_TABLE = {0x7B, 0x41, 0x42, 0x43, 0x44, 0x45, 0x46, 0x47, 0x48, 0x49};
private final static byte[] EBCDIC_CUST_NEG_TABLE = {0x7D, 0x4A, 0x4B, 0x4C, 0x4D, 0x4E, 0x4F, 0x50, 0x51, 0x52};
private final static byte[] EBCDIC_CUST_POS_SIGNS = {(byte)0xC0,(byte)0xC1,(byte)0xC2,(byte)0xC3,(byte)0xC4,(byte)0xC5,(byte)0xC6,(byte)0xC7,(byte)0xC8,(byte)0xC9};
private final static byte[] EBCDIC_CUST_NEG_SIGNS = {(byte)0xD0,(byte)0xD1,(byte)0xD2,(byte)0xD3,(byte)0xD4,(byte)0xD5,(byte)0xD6,(byte)0xD7,(byte)0xD8,(byte)0xD9};

// Long Double constants
private final static byte BYTE_SIGN_MASK = (byte)0x80;


/**
 * Converts the byte array to a char.  Assumes a 2-byte unicode character.
 * @return A char represented by the byte array.
 * @param sb Byte array containing a byte representation of a char.
 */

public static char convertFromByteToChar (byte[] sb )
{
   return ((char) convertFromByteToShort (sb));
}


/**
 * Converts the byte array to a double.
 * @return A double represented by the byte array.
 * @param sb Byte array containing a byte representation of a double.
 */

public static double convertFromByteToDouble (byte[] sb)
{
   long doubleLongBits = convertFromByteToLong(sb);

   return ( Double.longBitsToDouble(doubleLongBits));
}


/**
 * Converts the byte array to a float.
 * @return A float represented by the byte array.
 * @param sb Byte array containing a byte representation of a float.
 */

public static float convertFromByteToFloat (byte[] sb)
{
   int floatIntBits = convertFromByteToInt( sb);

   return ( Float.intBitsToFloat(floatIntBits) );
}


/**
 * Converts the byte array to an int.
 * @return An int represented by the byte array.
 * @param sb Byte array containing a byte representation of an int.
 */

public static int convertFromByteToInt (byte[] sb)
{
   int i =  (((int) sb[0]) << 24) >>> 0  |
            (((int) sb[1]) << 24) >>> 8  |
            (((int) sb[2]) << 24) >>> 16 |
            (((int) sb[3]) << 24) >>> 24;
   return i;
}


/**
 * Converts the byte array to a long.
 * @return A long represented by the byte array.
 * @param sb Byte array containing a byte representation of a long.
 */

public static long convertFromByteToLong (byte[] sb)
{
   long l = ((((long) sb[0]) << 56) & 0xFF00000000000000L) |
            ((((long) sb[1]) << 48) & 0x00FF000000000000L) |
            ((((long) sb[2]) << 40) & 0x0000FF0000000000L) |
            ((((long) sb[3]) << 32) & 0x000000FF00000000L) |
            ((((long) sb[4]) << 24) & 0x00000000FF000000L) |
            ((((long) sb[5]) << 16) & 0x0000000000FF0000L) |
            ((((long) sb[6]) << 8)  & 0x000000000000FF00L) |
            ((((long) sb[7]) )      & 0x00000000000000FFL);
   return l;
}


/**
 * Converts the byte array to a short.
 * @return A short represented by the byte array.
 * @param sb Byte array containing a byte representation of a short.
 */

public static short convertFromByteToShort (byte[] sb)
{
   // left shift to high bytes, then right shift with 0 extension
   // to correct position (to force zeros)

   int i;
   i =   (((int) sb[0]) << 24 ) >>>16 |
      (((int) sb[1]) << 24 ) >>> 24;

   return ((short)i);
}


/**
 * Returns a byte array representation of a char.
 * @return A byte array representation of a char.
 * @param sb The char to represent in the byte array.
 */

public static byte[] convertToByte (char sb)
{
   byte[] tb = new byte[2];
   tb[0] = (byte) (((short)sb & 0xFF00) >>>  8);
   tb[1] = (byte) (((short)sb & 0x00FF) );
   return tb;
}


/**
 * Returns a byte array representation of a double.
 * @return A byte array representation of a double.
 * @param sb The double to represent in the byte array.
 */

public static byte[] convertToByte (double sb)
{
   Double tempDouble = new Double(sb);
   long doubleLongBits = Double.doubleToLongBits (sb);
   byte[] tb = convertToByte (doubleLongBits);
   return tb;
}


/**
 * Returns a byte array representation of a float.
 * @return A byte array representation of a float.
 * @param sb The float to represent in the byte array.
 */

public static byte[] convertToByte (float sb)
{
   Float tempFloat = new Float(sb);
   int floatIntBits = Float.floatToIntBits (sb);

   byte[] tb = convertToByte (floatIntBits);
   return tb;
}


/**
 * Returns a byte array representation of an int.
 * @return A byte array representation of an int.
 * @param sb The int to represent in the byte array.
 */

public static byte[] convertToByte (int sb)
{
   byte[] tb = new byte[4];

   tb[0] = (byte) ((sb & 0xFF000000) >>> 24);
   tb[1] = (byte) ((sb & 0x00FF0000) >>> 16);
   tb[2] = (byte) ((sb & 0x0000FF00) >>>  8);
   tb[3] = (byte) ((sb & 0x000000FF) );

   return tb;
}


/**
 * Returns a byte array representation of a long.
 * @return A byte array representation of a long.
 * @param sb The long to represent in the byte array.
 */

public static byte[] convertToByte (long sb)
{
   byte[] tb = new byte[8];

   tb[0] = (byte) ((sb & 0xFF00000000000000L) >>> 56);
   tb[1] = (byte) ((sb & 0x00FF000000000000L) >>> 48);
   tb[2] = (byte) ((sb & 0x0000FF0000000000L) >>> 40);
   tb[3] = (byte) ((sb & 0x000000FF00000000L) >>> 32);
   tb[4] = (byte) ((sb & 0x00000000FF000000L) >>> 24);
   tb[5] = (byte) ((sb & 0x0000000000FF0000L) >>> 16);
   tb[6] = (byte) ((sb & 0x000000000000FF00L) >>>  8);
   tb[7] = (byte) ((sb & 0x00000000000000FFL));

   return tb;
}


/**
 * Returns a byte array representation of a short.
 * @return A byte array representation of a short.
 * @param sb The short to represent in the byte array.
 */

public static byte[] convertToByte (short sb)
{
   byte[] tb = new byte[2];

   tb[0] = (byte) ((sb & 0xFF00) >>>  8);
   tb[1] = (byte) ((sb & 0x00FF) );

   return tb;
}

// To convert from IEEE to S390 we use the following formula:
// let r = exponent % 4;  q = exponent / 4;
// if q == 0 then m.2^x = m.16^q
// if q != 0 then m.2^x = (m.2^(r-4)).16^(q+1)  for positive q,
//                      = (m.2^-r).16^q) for negative q

public static long doubleToS390LongBits(double ieeeDouble) throws IOException
{
   // Trace.entry(clsName,"doubleToS390LongBits");
   // Trace.trace(2,clsName,sccsid);

   // Get the bit pattern
   long ieeeLongBits = Double.doubleToLongBits(ieeeDouble);
   // Trace.trace(2,clsName,"IEEE double bit pattern= " + Long.toString(ieeeLongBits,16));

   // Test the sign bit (0 = positive, 1 = negative)
   boolean positive = ((ieeeLongBits & DOUBLE_SIGN_MASK) == 0);

   // Deal with zero straight away...
   if ((ieeeLongBits & 0x7fffffffffffffffL) == 0)
   {
      // + or - 0.0
      // Trace.exit(clsName,"doubleToS390LongBits (zero number)");
      return ieeeLongBits;
   }

   // Extract the exponent
   long exponent = ieeeLongBits & DOUBLE_EXPONENT_MASK;
   // shift right 52 bits to get exponent in least significant byte
   exponent = exponent >>> 52;
   // subtract the bias to get the true value
   exponent = exponent - DOUBLE_BIAS;

   // Extract the mantissa
   long mantissa = ieeeLongBits & DOUBLE_MANTISSA_MASK;

   // Now begin the conversion to S390

   long remainder = Math.abs(exponent) % 4;
   long quotient = Math.abs(exponent) / 4;

   long s390Exponent = quotient;
   if ((exponent > 0) && (remainder != 0))
   {
      s390Exponent = s390Exponent + 1;
   }

   // put the sign back in
   if (exponent < 0)
   {
      s390Exponent = -s390Exponent;
   }

   // Add the bias
   s390Exponent += S390_DOUBLE_BIAS;

   // Now adjust the mantissa part
   long s390Mantissa = mantissa;
   // for an exponent greater than -DOUBLE_BIAS, add in the implicit bit
   if (exponent > (-DOUBLE_BIAS))
   {
      s390Mantissa = s390Mantissa | DOUBLE_MANTISSA_MSB_MASK;
   }
   else
   {
      // there is no implicit bit, so the mantissa is one bit to the right of what
      // we would normally expect.  We need to fix this for S390
      s390Mantissa = s390Mantissa << 1;
   }

   // S390 Mantissa starts 4 bits left of ieee one. The first of these is implied in
   // IEEE so only shift 3 places
   s390Mantissa = s390Mantissa << 3;

   if (remainder > 0)
   {
      if (exponent > 0)
      {
         // the next two lines perform the (m.2^(r-4)) part of the conversion
         int shift_places = (int) (4 - remainder);
         s390Mantissa = s390Mantissa >>> shift_places;
      }
      else
      {
         // to avoid loss of precision when the exponent is at a minimum,
         // we may need to shift the mantissa four places left and decrease the
         // s390 exponent by one before shifting right
         if ((exponent == - (DOUBLE_BIAS)) && ((s390Mantissa & 0x00f0000000000000L) == 0))
         {
            s390Mantissa = s390Mantissa << 4;
            s390Exponent = s390Exponent - 1;
         }
         // the next two lines perform the m.2-r part of the conversion
         s390Mantissa = s390Mantissa >>> remainder;
      }
   }

   // An exponent of -DOUBLE_BIAS is the smallest that IEEE can do.  S390 has
   // a wider range, and hence may be able to normalise the mantissa more than
   // is possible for IEEE
   // Each shift left of four bits is equivalent to multiplying by 16,
   // so the exponent must be reduced by 1
   if (exponent == - (DOUBLE_BIAS))
   {
      while ((s390Mantissa != 0) && ((s390Mantissa & 0x00f0000000000000L) == 0))
      {
         s390Mantissa = s390Mantissa << 4;
         s390Exponent = s390Exponent - 1;
      }
   }

   // if the exponent is now > 127, we have an overflow since IEEE can handle larger numbers
   // than S390 can.
   if (s390Exponent > 127)
   {
      // Trace.trace(1,clsName,"Exponent = " + s390Exponent);
      // Trace.exit(clsName,"doubleToS390LongBits (exponent outside of range)");
      // throw new IOException("Number outside of range for double precision S/390 Float");
      throw new IOException(MarshallResource.instance().getString(MarshallResource.MARSHRT_FPNUM_OUTOF_RANGE, Double.toString(ieeeDouble)));
   }
   else if (s390Exponent < 0)
   {
      // the number is too small to represent, set it to zero
      // Trace.trace(1,clsName,"Number too small to represent, rounding to zero");
      // Trace.exit(clsName,"doubleToS390LongBits (exponent outside of range)");
      return 0L;
   }

   // Assemble the s390BitPattern
   long s390Double = 0L;
   long s390ExponentBits = s390Exponent & 0x000000000000007FL; // make sure we only deal with 7 bits
   // add the exponent
   s390Double = s390ExponentBits << 56; // shift to MSB
   // add the sign
   if (!positive)
   {
      s390Double = s390Double | DOUBLE_SIGN_MASK;
   }
   // add the mantissa
   s390Double = s390Double | s390Mantissa;

   // Trace.trace(2,clsName,"S390 bit pattern = " + Long.toString(s390Double,16));
   // Trace.exit(clsName,"doubleToS390LongBits");
   return s390Double;
}


/**
 * This method swaps the endian of a byte buffer.
 * @param buffer The byte buffer.
 */

public static void endianSwap (byte[] buffer)
{
   byte b;
   int leftSide = 0;
   int rightSide = buffer.length - 1;

   while (leftSide < rightSide)
   {
      b = buffer[leftSide];
      buffer[leftSide] = buffer[rightSide];
      buffer[rightSide] = b;
      leftSide++;
      rightSide--;
   }

   return;
}


/**
 * Returns the integer value of the formula string for an offset or size.
 * @return The integer value of the formula string for an offset or size.
 * @param formula The formula to be evaluated.
 */

public static int evaluateFormula (String formula)
   throws IllegalArgumentException
{
   if (formula == null)
      throw new IllegalArgumentException(MarshallResource.instance().getString(MarshallResource.MARSHRT_FORMULA_NULL));

   // Assumes for now that the string formula is a simple integer

   int value = Integer.parseInt (formula);

   return (value);
}


/**
 * Returns the integer value of the formula string for an offset or size.
 * @return The integer value of the formula string for an offset or size.
 * @param formula The formula to be evaluated.
 * @param valFieldNameMap A map of val() formulas to field names.
 * @param fh The format handler associated with the valFieldNameMap.
 * @throws IllegalArgumentException if the formula is null.
 *
 * The formula can be comprised of the following functions:
 * neg(x)   := -x       // prefix negate
 * add(x,y) := x+y      // infix add
 * sub(x,y) := x-y      // infix subtract
 * mpy(x,y) := x*y      // infix multiply
 * div(x,y) := x/y      // infix divide
 * max(x,y) := max(x,y)
 * min(x,y) := min(x,y)
 *
 * mod(x,y) := x mod y
 *
 * The mod function is defined as mod(x,y) = r where r is the smallest non-negative integer
 * such that x-r is evenly divisible by y. So mod(7,4) is 3, but mod(-7,4) is 1. If y is a
 * power of 2, then mod(x,y) is equal to the bitwise-and of x and y-1.
 *
 * val(1, m, n, o,..)
 *
 * The val function returns the value of a field in the model. The val function takes one
 * or more arguments, and the first argument refers to a level-1 field in the type model and must be either:
 *    - the name of a level-1 field described in the language model
 *    - the integer 1 (indicating that the level-1 parent of the current structure is meant)
 * If the first argument to the val function is the integer 1, then and only then are subsequent arguments
 * permitted. These subsequent arguments are integers that the specify the ordinal number within its
 * substructure of the subfield that should be dereferenced.
 */

public static int evaluateFormula (String formula, HashMap valFieldNameMap, WSIFFormatHandler fh)
   throws IllegalArgumentException
{
   if (formula == null)
      throw new IllegalArgumentException(MarshallResource.instance().getString(MarshallResource.MARSHRT_FORMULA_NULL));

   int result = 0;

   int index = formula.indexOf("(");

   if (index == -1)           // It's a number not an expression
   {
      try
      {
         result = Integer.parseInt (formula);
      }
      catch (Exception exc) { }

      return (result);
   }

   // Determine the outermost function
   String function = formula.substring (0, index);

   if (function.equalsIgnoreCase ("val"))
   {
      String fieldName = (String)valFieldNameMap.get (formula);
      if (fieldName == null)
         return (0);

      try
      {
         Number num = (Number)fh.getElement (fieldName, 0);

         if (num == null)     // Element does not exist try as a number
            result = Integer.parseInt (fieldName);
         else
            result = num.intValue();
      }
      catch (Exception exc) { }

      return (result);
   }
   else if (function.equalsIgnoreCase ("neg"))
   {
      // The new formula is the content between the brackets
      formula = formula.substring (index+1, formula.length()-1);
      result = -1 * evaluateFormula (formula, valFieldNameMap, fh);
      return (result);
   }
   else
   {
      // Get the contents between the outermost brackets
      formula = formula.substring (index+1, formula.length()-1);
      char[] formulaChars = formula.toCharArray();

      // Get the left side and the right side of the operation

      int brackets = 0;
      int i = 0;

      for ( ; i < formulaChars.length; i++)
      {
         if (formulaChars[i] == '(')
            brackets++;
         else if (formulaChars[i] == ')')
            brackets--;
         else if (formulaChars[i] == ',')
         {
            if (brackets == 0)
               break;
         }
      }

      String leftSide = "0";
      String rightSide = "0";

      leftSide = formula.substring (0, i);
      rightSide = formula.substring (i+1);

      if (function.equalsIgnoreCase ("add"))
         result = evaluateFormula (leftSide, valFieldNameMap, fh) + evaluateFormula (rightSide, valFieldNameMap, fh);
      else if (function.equalsIgnoreCase ("mpy"))
         result = evaluateFormula (leftSide, valFieldNameMap, fh) * evaluateFormula (rightSide, valFieldNameMap, fh);
      else if (function.equalsIgnoreCase ("sub"))
         result = evaluateFormula (leftSide, valFieldNameMap, fh) - evaluateFormula (rightSide, valFieldNameMap, fh);
      else if (function.equalsIgnoreCase ("div"))
         result = evaluateFormula (leftSide, valFieldNameMap, fh) / evaluateFormula (rightSide, valFieldNameMap, fh);
      else if (function.equalsIgnoreCase ("max"))
         result = Math.max (evaluateFormula (leftSide, valFieldNameMap, fh), evaluateFormula (rightSide, valFieldNameMap, fh));
      else if (function.equalsIgnoreCase ("min"))
         result = Math.min (evaluateFormula (leftSide, valFieldNameMap, fh),  evaluateFormula (rightSide, valFieldNameMap, fh));
      else if (function.equalsIgnoreCase ("mod"))
         result = evaluateFormula (leftSide, valFieldNameMap, fh) % evaluateFormula (rightSide, valFieldNameMap, fh);
   }

   return (result);
}


/**
 * Fills a byte array with the given fill byte. The number of components
 * filled is equal to the <code>length</code> argument, and the fill
 * bytes are filled at positions <code>offset</code> through
 * <code>offset+length-1</code>.
 * <p>
 * If any of the following is true, an
 * <code>ArrayIndexOutOfBoundsException</code> is
 * thrown and the destination is not modified:
 * <ul>
 * <li>The <code>offset</code> argument is negative.
 * <li><code>offset+length</code> is greater than
 *     <code>target.length</code>, the length of the target array.
 * </ul>
 * <p>
 *
 * @param      fillByte  the byte to be filled in.
 * @param      target    the destination array of bytes.
 * @param      offset    start position in the target array.
 * @param      length    the number of array elements to be filled.
 * @exception  ArrayIndexOutOfBoundsException  if filling would cause
 *               access of data outside array bounds.
 */

public static void fillBytes (byte fillByte, byte[] target, int offset, int length)
   throws ArrayIndexOutOfBoundsException
{
   if (length > 0)
   {
      int span = 1;
      target[offset] = fillByte;

      while (span * 2 <= length)
      {
         System.arraycopy (target, offset, target, offset + span, span);
         span = span * 2;
      }

      System.arraycopy (target, offset, target, offset + span, length - span);
   }
}


// To convert from IEEE to S390 we use the following formula:
// let r = exponent % 4;  q = exponent / 4;
// if q == 0 then m.2^x = m.16^q
// if q != 0 then m.2^x = (m.2^(r-4)).16^(q+1)  for positive q,
//                      = (m.2^-r).16^q) for negative q

public static int floatToS390IntBits(float ieeeFloat) throws IOException
{
   // Trace.entry(clsName,"floatToS390IntBits");
   // Trace.trace(2,clsName,sccsid);

   // Get the bit pattern
   int ieeeIntBits = Float.floatToIntBits(ieeeFloat);
   // Trace.trace(2,clsName,"IEEE bit pattern = " + Integer.toString(ieeeIntBits,16));

   // Test the sign bit (0 = positive, 1 = negative)
   boolean positive = ((ieeeIntBits & FLOAT_SIGN_MASK) == 0);

   // Deal with zero straight away...
   if ((ieeeIntBits & 0x7fffffff) == 0)
   {
      // + or - 0.0
      // Trace.exit(clsName,"floatToS390IntBits (zero number)");
      return ieeeIntBits;
   }

   // Extract the exponent
   int exponent = ieeeIntBits & FLOAT_EXPONENT_MASK;
   // shift right 23 bits to get exponent in least significant byte
   exponent = exponent >>> 23;
   // subtract the bias to get the true value
   exponent = exponent - FLOAT_BIAS;

   // Extract the mantissa
   int mantissa = ieeeIntBits & FLOAT_MANTISSA_MASK;
   // for an exponent greater than -FLOAT_BIAS, add in the implicit bit
   if (exponent > (-FLOAT_BIAS))
   {
      mantissa = mantissa | FLOAT_MANTISSA_MSB_MASK;
   }

   // Now begin the conversion to S390

   int remainder = Math.abs(exponent) % 4;
   int quotient = Math.abs(exponent) / 4;

   int s390Exponent = quotient;
   if ((exponent > 0) && (remainder != 0))
   {
      s390Exponent = s390Exponent + 1;
   }

   // put the sign back in
   if (exponent < 0)
   {
      s390Exponent = -s390Exponent;
   }

   // Add the bias
   s390Exponent += S390_FLOAT_BIAS;

   // Now adjust the mantissa part
   int s390Mantissa = mantissa;
   if (remainder > 0)
   {
      if (exponent > 0)
      {
         // the next two lines perform the (m.2^(r-4)) part of the conversion
         int shift_places = 4 - remainder;
         s390Mantissa = s390Mantissa >>> shift_places;
      }
      else
      {
         // to avoid loss of precision when the exponent is at a minimum,
         // we may need to shift the mantissa four places left, and decrease the
         // s390Exponent by one before shifting right
         if ((exponent == - (FLOAT_BIAS)) && ((s390Mantissa & 0x00f00000) == 0))
         {
            s390Mantissa = s390Mantissa << 4;
            s390Exponent = s390Exponent - 1;
         }
         // the next two line perform the (m.2^-r) part of the conversion
         int shift_places = remainder;
         s390Mantissa = s390Mantissa >>> shift_places;
      }
   }

   // An exponent of -FLOAT_BIAS is the smallest that IEEE can do.  S390 has
   // a wider range, and hence may be able to normalise the mantissa more than
   // is possible for IEEE
   // Also, since an exponent of -FLOAT_BIAS has no implicit bit set, the mantissa
   // starts with a value of 2^-1 at the second bit of the second byte.  We thus need
   // to shift one place left to move the mantissa to the S390 position
   // Follwoing that, we notmalise as follows:
   // Each shift left of four bits is equivalent to multiplying by 16,
   // so the exponent must be reduced by 1
   if (exponent == - (FLOAT_BIAS))
   {
      s390Mantissa = s390Mantissa << 1;
      while ((s390Mantissa != 0) && ((s390Mantissa & 0x00f00000) == 0))
      {
         s390Mantissa = s390Mantissa << 4;
         s390Exponent = s390Exponent - 1;
      }
   }

   // Assemble the s390BitPattern
   int s390Float = 0;
   int s390ExponentBits = s390Exponent & 0x0000007F; // make sure we only deal with 7 bits
   // add the exponent
   s390Float = s390ExponentBits << 24; // shift to MSB
   // add the sign
   if (!positive)
   {
      s390Float = s390Float | FLOAT_SIGN_MASK;
   }
   // add the mantissa
   s390Float = s390Float | s390Mantissa;

   // Trace.trace(2,clsName,"S390 Bit pattern = " + Integer.toString(s390Float,16));
   // Trace.exit(clsName,"floatToS390IntBits");
   return s390Float;
}


// To convert from S390 to IEEE we use the fomula:
// m.16^x = m.2^4x, and then normalise by shifting the mantissa up to three
// places left

public static float intS390BitsToFloat(int floatBits) throws IOException
{
   // Trace.entry(clsName,"intS390BitsToFloat");
   // Trace.trace(2,clsName,sccsid);
   // Trace.trace(2,clsName,"S390 Bit pattern = " + Integer.toString(floatBits,16));

   // Test the sign bit (0 = positive, 1 = negative)
   boolean positive = ((floatBits & FLOAT_SIGN_MASK) == 0);

   // Deal with zero straight away...
   if ((floatBits & 0x7fffffff) == 0)
   {
      // + or - 0.0
      // Trace.exit(clsName,"intS390BitsToFloat (zero number)");
      if (positive)
      {
         return 0.0F;
      }
      else
      {
         return - (0.0F);
      }
   }

   int mantissa = floatBits & S390_FLOAT_MANTISSA_MASK;
   int exponent = floatBits & S390_FLOAT_EXPONENT_MASK;

   // move the exponent into the LSB
   exponent = exponent >> 24;
   // subtract the bias
   exponent = exponent - S390_FLOAT_BIAS;

   // caculate the IEEE exponent
   int ieeeExponent = exponent * 4;

   // Normalise the mantissa
   int ieeeMantissa = mantissa;
   // Deal with exponents <= -FLOAT_BIAS
   if (ieeeExponent <= - (FLOAT_BIAS))
   {
      // ieeeMantissa is one place to the right since there is no implicit bit set
      ieeeMantissa = ieeeMantissa >> 1;
      // now increase the exponent until it reaches -FLOAT_BIAS, shifting right one
      // place at each stage to compensate
      while (ieeeExponent < - (FLOAT_BIAS))
      {
         ieeeExponent = ieeeExponent + 1;
         ieeeMantissa = ieeeMantissa >> 1;
      }
   }

   // Deal with exponents greater than -FLOAT_BIAS
   while ((ieeeMantissa != 0) && ((ieeeMantissa & FLOAT_MANTISSA_MSB_MASK) == 0) && (ieeeExponent > - (FLOAT_BIAS)))
   {
      ieeeMantissa = ieeeMantissa << 1; // *2
      ieeeExponent = ieeeExponent - 1; // /2
   }

   // s390 has a wider range than IEEE, so deal with over and underflows
   if (ieeeExponent < -149)
   {
      // Trace.exit(clsName,"intS390BitsToFloat (underflow, returning zero)");
      return 0.0F; // underflow
   }
   else if (ieeeExponent > 128)
   {
      if (positive)
      {
         // Trace.exit(clsName,"intS390BitsToFloat (overflow, returning +INFINITY)");
         return (Float.MAX_VALUE * 2); // + infinity
      }
      else
      {
         // Trace.exit(clsName,"intS390BitsToFloat (overflow, returning -INFINITY)");
         return - (Float.MAX_VALUE * 2); // -infinity
      }
   }

   // Build the IEEE float
   int ieeeBits = 0;
   if (!positive)
   {
      ieeeBits = ieeeBits | FLOAT_SIGN_MASK;
   }

   // add the bias to the exponent
   ieeeExponent = ieeeExponent + FLOAT_BIAS;
   // move it to the IEEE exponent position
   ieeeExponent = ieeeExponent << 23;
   // add to the result
   ieeeBits = ieeeBits | ieeeExponent;

   // mask the top bit of the mantissa (implicit in IEEE)
   ieeeMantissa = ieeeMantissa & FLOAT_MANTISSA_MASK;
   // add to the result
   ieeeBits = ieeeBits | ieeeMantissa;

   // Trace.trace(2,clsName,"IEEE Bit pattern = " + Integer.toString(ieeeBits,16));
   // Trace.exit(clsName,"intS390BitsToFloat");
   return Float.intBitsToFloat(ieeeBits);
}


// To convert from S390 to IEEE we use the fomula:
// m.16^x = m.2^4x, and then normalise by shifting the mantissa up to three
// places left

public static double longS390BitsToDouble(long doubleBits) throws IOException
{
   // Trace.entry(clsName,"longS390BitsToDouble");
   // Trace.trace(2,clsName,sccsid);
   // Trace.trace(2,clsName,"S390 bit pattern = " + Long.toString(doubleBits,16));

   // Test the sign bit (0 = positive, 1 = negative)
   boolean positive = ((doubleBits & DOUBLE_SIGN_MASK) == 0);

   // Deal with zero straight away...
   if ((doubleBits & 0x7fffffffffffffffL) == 0)
   {
      // + or - 0.0
      // Trace.exit(clsName,"longS390BitsToDouble (zero number)");;
      if (positive)
      {
         return 0.0D;
      }
      else
      {
         return - (0.0D);
      }
   }

   long mantissa = doubleBits & S390_DOUBLE_MANTISSA_MASK;
   long exponent = doubleBits & S390_DOUBLE_EXPONENT_MASK;

   // move the exponent into the LSB
   exponent = exponent >> 56;
   // subtract the bias
   exponent = exponent - S390_DOUBLE_BIAS;

   // caculate the IEEE exponent
   long ieeeExponent = exponent * 4;

   // Normalise the mantissa
   long ieeeMantissa = mantissa;
   // IEEE mantissa starts three places right of S390 (+ implicit bit)
   ieeeMantissa = ieeeMantissa >> 3;
   // if this is the samllest possible exponent, then there is no implicit bit,
   // and so we need to shift an extra bit
   if (ieeeExponent <= - (DOUBLE_BIAS))
   {
      ieeeMantissa = ieeeMantissa >> 1;
      // now increase the exponent until it reaches -DOUBLE_BIAS, shifting
      // right one place at each stage to compensate
      while (ieeeExponent < - (DOUBLE_BIAS))
      {
         ieeeExponent = ieeeExponent + 1;
         ieeeMantissa = ieeeMantissa >> 1;
      }
   }

   // complete the normalisation for exponents > -DOUBLE_BIAS
   while ((ieeeMantissa != 0) && ((ieeeMantissa & DOUBLE_MANTISSA_MSB_MASK) == 0) && (ieeeExponent > - (DOUBLE_BIAS)))
   {
      ieeeMantissa = ieeeMantissa << 1; // *2
      ieeeExponent = ieeeExponent - 1; // /2
   }

   // s390 has a wider range than IEEE, so deal with over and underflows
   if (ieeeExponent < -1045)
   {
      return 0.0F; // underflow
   }
   else if (ieeeExponent > 1024)
   {
      if (positive)
      {
         // Trace.trace(1,clsName,"overflow - returning +INFINITY");
         // Trace.exit(clsName,"longS90BitsToDouble");
         return (Double.MAX_VALUE * 2); // + infinity
      }
      else
      {
         // Trace.trace(1,clsName,"overflow - returning -INFINITY");
         // Trace.exit(clsName,"longS390BitsToDouble");
         return - (Double.MAX_VALUE * 2); // -infinity
      }
   }

   // Built the IEEE double
   long ieeeBits = 0;
   if (!positive)
   {
      ieeeBits = ieeeBits | DOUBLE_SIGN_MASK;
   }

   // add the bias to the exponent
   ieeeExponent = ieeeExponent + DOUBLE_BIAS;
   // move it to the IEEE exponent position
   ieeeExponent = ieeeExponent << 52;
   // add to the result
   ieeeBits = ieeeBits | ieeeExponent;

   // mask the top bit of the mantissa (implicit in IEEE)
   ieeeMantissa = ieeeMantissa & DOUBLE_MANTISSA_MASK;
   // add to the result
   ieeeBits = ieeeBits | ieeeMantissa;

   // Trace.trace(2,clsName,"IEEE bit pattern = " + Long.toString(ieeeBits,16));
   // Trace.exit(clsName,"longS390BitsToDouble");
   return Double.longBitsToDouble(ieeeBits);
}

public static BigDecimal convertPackedDecimalToBigDecimal (byte[] packedDecimalBytes)
{
   boolean positive = true;
   StringBuffer sb = new StringBuffer();
   int index = packedDecimalBytes.length-1;

   sb.append (Character.forDigit(((packedDecimalBytes[index] >>> 4) & 0x0F), 16));

   index --;

   for (; index >=0; index--)
   {
      // Get low bit digit
      sb.append (Character.forDigit( (packedDecimalBytes[index] & 0x0F), 16));
      // Get high bit digit
      sb.append (Character.forDigit((( packedDecimalBytes[index] >>> 4) & 0x0F),16));
   }

   if ((packedDecimalBytes[(packedDecimalBytes.length-1)] & 0x0F) == PACKED_DEC_NEGOR)
      sb.append ("-");

   sb = sb.reverse();

   BigDecimal retVal = new BigDecimal(sb.toString());

   return (retVal);
}

// Packed decimal format is as follows:  Each nibble (4 bits) represents one digit (0-9).
// The most significant digit is on the left.  The least significant nibble (rightmost) is
// a sign bit.  A value of 0x0C (1100) is positive, 0x0D (1101) is negative, 0x0F is unsigned.

public static long convertPackedDecimalToBinary (byte[] packedDecimalBytes)
   throws IllegalArgumentException
{
   long retVal = 0;
   long multiplier = 1;
   boolean positive = true;

   // Work backwards from the least significant byte, each byte holds two digits...
   for (int b = (packedDecimalBytes.length - 1); b >= 0; b--)
   {
      byte thisByte = packedDecimalBytes[b];
      int leastSignificantDigit = thisByte & 0x0F;
      int mostSignificantDigit = (thisByte >>> 4) & 0x0F;
      // >>> = shift right with zero extension

      // Special case the rightmost byte, whose lower nibble indicates sign
      if (b == (packedDecimalBytes.length - 1))
      {
         switch (leastSignificantDigit)
         {
            case PACKED_DEC_POSOR :
               positive = true;
               break;
            case PACKED_DEC_NEGOR :
               positive = false;
               break;
            case PACKED_DEC_NOSIGN :
               // unsigned = positive
               positive = true;
               break;

            default :
               throw new IllegalArgumentException(MarshallResource.instance().getString(MarshallResource.MARSHRT_PACKDEC_SIGN_NOT_VALID, Integer.toString(leastSignificantDigit)));
         }

         // "Move" the high nibble into the lower one, since for the leftmost
         // byte this nibble is units, not tens.
         leastSignificantDigit = mostSignificantDigit;
         mostSignificantDigit = 0;
      }

      // sanity check...
      if ((leastSignificantDigit > 9) || (mostSignificantDigit > 9))
      {
         throw new IllegalArgumentException(MarshallResource.instance().getString(MarshallResource.MARSHRT_PDNUM_OUTOF_RANGE, Integer.toString(thisByte)));
      }

      int binaryValue = leastSignificantDigit + (10 * mostSignificantDigit);
      retVal = retVal + multiplier * binaryValue;

      if (b == (packedDecimalBytes.length - 1))
      {
         multiplier = multiplier * 10;
      }
      else
      {
         multiplier = multiplier * 100;
      }
   }

   if (!positive)
   {
      retVal = -retVal;
   }

   return (retVal);
}

public static byte[] convertToPackedDecimal (char[] pdChars, int size, boolean signed)
{
   byte[] retVal = new byte[size];
   int rvIndex = size - 1;
   int charIndex = pdChars.length - 1;
   int charEnd = 0;

   byte signByte = (byte)(Character.digit(pdChars[charIndex], 16));

   if (signed)
   {
      if (pdChars[0] == '-')
         signByte = (byte)((signByte << 4) | PACKED_DEC_NEGOR);
      else
         signByte = (byte)((signByte << 4) | PACKED_DEC_POSOR);
   }
   else
      signByte = (byte)((signByte << 4) | PACKED_DEC_NOSIGN);

   if ((pdChars[0] == '-') || (pdChars[0] == '+'))
      charEnd = 1;

   retVal[rvIndex] = signByte;
   rvIndex--;
   charIndex--;

   int i;
   for (i = rvIndex; ((i >= 0) && (charIndex >= charEnd)); i--)
   {
      int leastSignificantDigit = Character.digit(pdChars[charIndex], 16);
      int mostSignificantDigit = 0;
      charIndex--;
      if (charIndex >= charEnd)
         mostSignificantDigit = Character.digit(pdChars[charIndex], 16);
      retVal[i] = (byte) (leastSignificantDigit | (mostSignificantDigit << 4));
      charIndex--;
   }

   return (retVal);
}


public static byte[] convertToPackedDecimal (long binary, int size, boolean signed)
{
   byte[] retVal = new byte[size];
   boolean positive = (binary >= 0);
   if ((!signed) && (!positive))
      binary = -binary;

   // Deal with the rightmost byte specially, since it contains the sign bit
   int decimalDigit = (int) Math.abs (binary % 10);
   binary = (long) (binary / 10);
   byte signByte;

   if (signed)
   {
      if (positive)
         signByte = (byte) ((decimalDigit << 4) | PACKED_DEC_POSOR);
      else
         signByte = (byte) ((decimalDigit << 4) | PACKED_DEC_NEGOR);
   }
   else
      signByte = (byte) ((decimalDigit << 4) | PACKED_DEC_NOSIGN);

   retVal[size-1] = signByte;

   for (int byteIndex = (size-2); ((byteIndex >= 0) && (binary != 0)); byteIndex--)
   {
      int leastSignificantDigit = (int) Math.abs (binary % 10);
      binary = binary / 10;
      int mostSignificantDigit = (int) Math.abs (binary % 10);
      binary = binary / 10;
      retVal[byteIndex] = (byte) (leastSignificantDigit | (mostSignificantDigit << 4));
   }

   return (retVal);
}


private static byte[] getPosTable (String codePage)
{
	byte[] result = EBCDIC_CUST_POS_TABLE;

	if (codePage == null)
		return (result);

	try
	{
		String str = new String(EBCDIC_CUST_POS_SIGNS, codePage);
		result = str.getBytes ("8859_1");
	}
	catch (Exception exc)
	{
		result = EBCDIC_CUST_POS_TABLE;
	}

	return (result);
}


private static byte[] getNegTable (String codePage)
{
	byte[] result = EBCDIC_CUST_NEG_TABLE;

	if (codePage == null)
		return (result);

	try
	{
		String str = new String(EBCDIC_CUST_NEG_SIGNS, codePage);
		result = str.getBytes ("8859_1");
	}
	catch (Exception exc)
	{
		result = EBCDIC_CUST_NEG_TABLE;
	}

	return (result);
}


public static BigDecimal convertExternalDecimalToBigDecimal (byte[] externalDecimalBytes, boolean signed, int signFormat, int externalDecimalSign)
	throws IllegalArgumentException
{
	return (convertExternalDecimalToBigDecimal (externalDecimalBytes, signed, signFormat, externalDecimalSign, null));
}


public static BigDecimal convertExternalDecimalToBigDecimal (byte[] externalDecimalBytes, boolean signed, int signFormat, int externalDecimalSign, String codePage)
   throws IllegalArgumentException
{
   long multiplier = 1;
   boolean isPositive = true;
   byte posor, negor, minus;
   StringBuffer sb = new StringBuffer();
   byte[] posTable = null;
   byte[] negTable = null;

   switch (externalDecimalSign)
   {
      case MarshallExternalDecimalUtils.EXTERNAL_DECIMAL_SIGN_EBCDIC:
         posor = EBCDIC_EXTDEC_POSOR;
         negor = EBCDIC_EXTDEC_NEGOR;
         minus = EBCDIC_MINUS;
         break;

      case MarshallExternalDecimalUtils.EXTERNAL_DECIMAL_SIGN_EBCDIC_CUSTOM:
         posTable = getPosTable (codePage);
         negTable = getNegTable (codePage);
         posor = ASCII_EXTDEC_POSOR;
         negor = ASCII_EXTDEC_NEGOR;
         minus = EBCDIC_CUST_MINUS;
         break;

      case MarshallExternalDecimalUtils.EXTERNAL_DECIMAL_SIGN_ASCII:
         posor = ASCII_EXTDEC_POSOR;
         negor = ASCII_EXTDEC_NEGOR;
         minus = ASCII_MINUS;
         break;

      default:
         throw new IllegalArgumentException(MarshallResource.instance().getString(MarshallResource.MARSHRT_EXTDEC_SIGN_NOT_VALID, Integer.toString(externalDecimalSign)));
   }

   int start = externalDecimalBytes.length - 1;
   int end = 0;

   if (signed)    // Need to extract positive or negative sign
   {
      switch (signFormat)
      {
         case MarshallExternalDecimalUtils.SIGN_FORMAT_LEADING:
            if (externalDecimalSign != MarshallExternalDecimalUtils.EXTERNAL_DECIMAL_SIGN_EBCDIC_CUSTOM)
            {
               if ((byte)(externalDecimalBytes[0] & 0xF0) == negor)
                  isPositive = false;
            }
            else
            {
               // Find the value is in the table and convert it to the ascii equivalent
               for (int i = 0; i < negTable.length; i++)
               {
                  if (externalDecimalBytes[0] == posTable[i])
                  {
                     externalDecimalBytes[0] = (byte)((i & 0x0f) | posor);
                     break;
                  }
                  else if (externalDecimalBytes[0] == negTable[i])
                  {
                     externalDecimalBytes[0] = (byte)((i & 0x0f) | negor);
                     isPositive = false;
                     break;
                  }
               }
            }
            break;

         case MarshallExternalDecimalUtils.SIGN_FORMAT_LEADING_SEPARATE:
            if (externalDecimalBytes[0] == minus)
               isPositive = false;
            end = 1;
            break;

         case MarshallExternalDecimalUtils.SIGN_FORMAT_TRAILING:
            if (externalDecimalSign != MarshallExternalDecimalUtils.EXTERNAL_DECIMAL_SIGN_EBCDIC_CUSTOM)
            {
               if ((byte)(externalDecimalBytes[externalDecimalBytes.length - 1] & 0xF0) == negor)
                  isPositive = false;
            }
            else
            {
               // Find the value is in the table and convert it to the ascii equivalent
               for (int i = 0; i < negTable.length; i++)
               {
                  if (externalDecimalBytes[externalDecimalBytes.length - 1] == posTable[i])
                  {
                     externalDecimalBytes[externalDecimalBytes.length-1] = (byte)((i & 0x0f) | posor);
                     break;
                  }
                  else if (externalDecimalBytes[externalDecimalBytes.length - 1] == negTable[i])
                  {
                     externalDecimalBytes[externalDecimalBytes.length-1] = (byte)((i & 0x0f) | negor);
                     isPositive = false;
                     break;
                  }
               }
            }
            break;

         case MarshallExternalDecimalUtils.SIGN_FORMAT_TRAILING_SEPARATE:
            if (externalDecimalBytes[externalDecimalBytes.length - 1] == minus)
               isPositive = false;
            start = externalDecimalBytes.length - 2;
            break;

         default:
            throw new IllegalArgumentException(MarshallResource.instance().getString(MarshallResource.MARSHRT_EXTDEC_SIGN_FMT_NOT_VALID, Integer.toString(signFormat)));
      }
   }

   // Work backwards from the least significant byte
   for (int i = start; i >= end; i--)
      sb.append (Character.forDigit((externalDecimalBytes[i] & 0x0F), 16));

   if (!isPositive)
      sb.append ("-");

   sb = sb.reverse();

   BigDecimal retval = new BigDecimal(sb.toString());
   return (retval);
}


public static long convertExternalDecimalToBinary (byte[] externalDecimalBytes, boolean signed, int signFormat, int externalDecimalSign)
	throws IllegalArgumentException
{
	return (convertExternalDecimalToBinary (externalDecimalBytes, signed, signFormat, externalDecimalSign, null));
}


public static long convertExternalDecimalToBinary (byte[] externalDecimalBytes, boolean signed, int signFormat, int externalDecimalSign, String codePage)
   throws IllegalArgumentException
{
   long result = 0;
   long multiplier = 1;
   boolean isPositive = true;
   byte posor, negor, minus;
   byte[] posTable = null;
   byte[] negTable = null;

   switch (externalDecimalSign)
   {
      case MarshallExternalDecimalUtils.EXTERNAL_DECIMAL_SIGN_EBCDIC:
         posor = EBCDIC_EXTDEC_POSOR;
         negor = EBCDIC_EXTDEC_NEGOR;
         minus = EBCDIC_MINUS;
         break;

      case MarshallExternalDecimalUtils.EXTERNAL_DECIMAL_SIGN_EBCDIC_CUSTOM:
         posTable = getPosTable (codePage);
         negTable = getNegTable (codePage);
         posor = ASCII_EXTDEC_POSOR;
         negor = ASCII_EXTDEC_NEGOR;
         minus = EBCDIC_CUST_MINUS;
         break;

      case MarshallExternalDecimalUtils.EXTERNAL_DECIMAL_SIGN_ASCII:
         posor = ASCII_EXTDEC_POSOR;
         negor = ASCII_EXTDEC_NEGOR;
         minus = ASCII_MINUS;
         break;

      default:
      throw new IllegalArgumentException(MarshallResource.instance().getString(MarshallResource.MARSHRT_EXTDEC_SIGN_NOT_VALID, Integer.toString(externalDecimalSign)));
   }

   int start = externalDecimalBytes.length - 1;
   int end = 0;

   if (signed)    // Need to extract positive or negative sign
   {
      switch (signFormat)
      {
         case MarshallExternalDecimalUtils.SIGN_FORMAT_LEADING:
            if (externalDecimalSign != MarshallExternalDecimalUtils.EXTERNAL_DECIMAL_SIGN_EBCDIC_CUSTOM)
            {
               if ((byte)(externalDecimalBytes[0] & 0xF0) == negor)
                  isPositive = false;
            }
            else
            {
               // Find the value is in the table and convert it to the ascii equivalent
               for (int i = 0; i < negTable.length; i++)
               {
                  if (externalDecimalBytes[0] == posTable[i])
                  {
                     externalDecimalBytes[0] = (byte)((i & 0x0f) | posor);
                     break;
                  }
                  else if (externalDecimalBytes[0] == negTable[i])
                  {
                     externalDecimalBytes[0] = (byte)((i & 0x0f) | negor);
                     isPositive = false;
                     break;
                  }
               }
            }
            break;

         case MarshallExternalDecimalUtils.SIGN_FORMAT_LEADING_SEPARATE:
            if (externalDecimalBytes[0] == minus)
               isPositive = false;
            end = 1;
            break;

         case MarshallExternalDecimalUtils.SIGN_FORMAT_TRAILING:
            if (externalDecimalSign != MarshallExternalDecimalUtils.EXTERNAL_DECIMAL_SIGN_EBCDIC_CUSTOM)
            {
               if ((byte)(externalDecimalBytes[externalDecimalBytes.length - 1] & 0xF0) == negor)
                  isPositive = false;
            }
            else
            {
               // Find the value is in the table and convert it to the ascii equivalent
               for (int i = 0; i < negTable.length; i++)
               {
                  if (externalDecimalBytes[externalDecimalBytes.length - 1] == posTable[i])
                  {
                     externalDecimalBytes[externalDecimalBytes.length-1] = (byte)((i & 0x0f) | posor);
                     break;
                  }
                  else if (externalDecimalBytes[externalDecimalBytes.length - 1] == negTable[i])
                  {
                     externalDecimalBytes[externalDecimalBytes.length-1] = (byte)((i & 0x0f) | negor);
                     isPositive = false;
                     break;
                  }
               }
            }
            break;

         case MarshallExternalDecimalUtils.SIGN_FORMAT_TRAILING_SEPARATE:
            if (externalDecimalBytes[externalDecimalBytes.length - 1] == minus)
               isPositive = false;
            start = externalDecimalBytes.length - 2;
            break;

         default:
            throw new IllegalArgumentException(MarshallResource.instance().getString(MarshallResource.MARSHRT_EXTDEC_SIGN_FMT_NOT_VALID, Integer.toString(signFormat)));
      }
   }

   // Work backwards from the least significant byte
   for (int i = start; i >= end; i--)
   {
      int binaryValue = externalDecimalBytes[i] & 0x0F;
      result += multiplier * binaryValue;
      multiplier *= 10;
   }

   if (!isPositive)
      result = -result;

   return (result);
}


public static byte[] convertToExternalDecimal (char[] extdChars, int size, boolean signed, int signFormat, int externalDecimalSign)
	throws IllegalArgumentException
{
	return (convertToExternalDecimal (extdChars, size, signed, signFormat, externalDecimalSign, null));
}


public static byte[] convertToExternalDecimal (char[] extdChars, int size, boolean signed, int signFormat, int externalDecimalSign, String codePage)
   throws IllegalArgumentException
{
   byte[] retVal = new byte[size];

   boolean isPositive = true;
   int charEnd = 0;
   byte[] posTable = null;
   byte[] negTable = null;

   if (extdChars[0] == '-')
      isPositive = false;

   if ((extdChars[0] == '-') || (extdChars[0] == '+'))
      charEnd++;

   byte zeroPad, posor, negor, plus, minus;

   switch (externalDecimalSign)
   {
      case MarshallExternalDecimalUtils.EXTERNAL_DECIMAL_SIGN_EBCDIC:
         zeroPad = EBCDIC_ZERO_PAD;
         posor =  EBCDIC_EXTDEC_POSOR;
         negor = EBCDIC_EXTDEC_NEGOR;
         plus = EBCDIC_PLUS;
         minus = EBCDIC_MINUS;
         break;

      case MarshallExternalDecimalUtils.EXTERNAL_DECIMAL_SIGN_EBCDIC_CUSTOM:
         posTable = getPosTable (codePage);
         negTable = getNegTable (codePage);
         zeroPad = EBCDIC_CUST_ZERO_PAD;
         posor =  EBCDIC_CUST_EXTDEC_POSOR;
         negor = EBCDIC_CUST_EXTDEC_NEGOR;
         plus = EBCDIC_CUST_PLUS;
         minus = EBCDIC_CUST_MINUS;
         break;

      case MarshallExternalDecimalUtils.EXTERNAL_DECIMAL_SIGN_ASCII:
         zeroPad = ASCII_ZERO_PAD;
         posor = ASCII_EXTDEC_POSOR;
         negor = ASCII_EXTDEC_NEGOR;
         plus = ASCII_PLUS;
         minus = ASCII_MINUS;
         break;

      default:
      throw new IllegalArgumentException(MarshallResource.instance().getString(MarshallResource.MARSHRT_EXTDEC_SIGN_NOT_VALID, Integer.toString(externalDecimalSign)));
   }

   int i = size - 1;
   int charIndex = extdChars.length - 1;

   if (signed && (signFormat == MarshallExternalDecimalUtils.SIGN_FORMAT_TRAILING_SEPARATE))    // Leave last byte for sign
      i = size - 2;

   // Fill the byte array with the numerical values or'd with 0xF0 (mvs) or 0x30 (ascii)
   for (; (i >= 0 && (charIndex >= charEnd)); i--, charIndex--)
   {
      // Get the rightmost digit
      int decimalDigit = Character.digit(extdChars[charIndex], 16);
      retVal[i] = (byte)(decimalDigit | zeroPad);
   }

   if (i >= 0)    // If the value is less than the size of the field
   {
      // Pad out start of number
      for (int j = i; j >=0; j--)
         retVal[j] = zeroPad;
   }

   if (signed)    // Determine where the sign should be in the buffer
   {
      switch (signFormat)
      {
         case MarshallExternalDecimalUtils.SIGN_FORMAT_LEADING:
            if (externalDecimalSign != MarshallExternalDecimalUtils.EXTERNAL_DECIMAL_SIGN_EBCDIC_CUSTOM)
            {
               if (isPositive)
                  retVal[0] = (byte)((retVal[0] & 0x0F) | posor);
               else
                  retVal[0] = (byte)((retVal[0] & 0x0F) | negor);
            }
            else
            {
               int index = retVal[0] & 0x0F; // Get the value of the digit and look up the value in the table
               if (isPositive)
                  retVal[0] = posTable[index];
               else
                  retVal[0] = negTable[index];
            }

            break;

         case MarshallExternalDecimalUtils.SIGN_FORMAT_LEADING_SEPARATE:
            if (isPositive)
               retVal[0] = plus;
            else
               retVal[0] = minus;

            break;

         case MarshallExternalDecimalUtils.SIGN_FORMAT_TRAILING:
            if (externalDecimalSign != MarshallExternalDecimalUtils.EXTERNAL_DECIMAL_SIGN_EBCDIC_CUSTOM)
            {
               if (isPositive)
                  retVal[size-1] = (byte)((retVal[size-1] & 0x0F) | posor);
               else
                  retVal[size-1] = (byte)((retVal[size-1] & 0x0F) | negor);
            }
            else
            {
               int index = retVal[size-1] & 0x0F;  // Get the value of the digit and look up the value in the table
               if (isPositive)
                  retVal[size-1] = posTable[index];
               else
                  retVal[size-1] = negTable[index];
            }

            break;

         case MarshallExternalDecimalUtils.SIGN_FORMAT_TRAILING_SEPARATE:
            if (isPositive)
               retVal[size-1] = plus;
            else
               retVal[size-1] = minus;

            break;

         default:
            throw new IllegalArgumentException(MarshallResource.instance().getString(MarshallResource.MARSHRT_EXTDEC_SIGN_FMT_NOT_VALID, Integer.toString(signFormat)));
      }
   }

   return (retVal);
}


public static byte[] convertToExternalDecimal (long binary, int size, boolean signed, int signFormat, int externalDecimalSign)
	throws IllegalArgumentException
{
	return (convertToExternalDecimal (binary, size, signed, signFormat, externalDecimalSign, null));
}


public static byte[] convertToExternalDecimal (long binary, int size, boolean signed, int signFormat, int externalDecimalSign, String codePage)
   throws IllegalArgumentException
{
   byte[] retVal = new byte[size];

   boolean isPositive = true;
   byte[] posTable = null;
   byte[] negTable = null;

   if (binary < 0)
   {
      isPositive = false;
      if (!signed)
         binary = -binary;
   }

   byte zeroPad, posor, negor, plus, minus;

   switch (externalDecimalSign)
   {
      case MarshallExternalDecimalUtils.EXTERNAL_DECIMAL_SIGN_EBCDIC:
         zeroPad = EBCDIC_ZERO_PAD;
         posor =  EBCDIC_EXTDEC_POSOR;
         negor = EBCDIC_EXTDEC_NEGOR;
         plus = EBCDIC_PLUS;
         minus = EBCDIC_MINUS;
         break;

      case MarshallExternalDecimalUtils.EXTERNAL_DECIMAL_SIGN_EBCDIC_CUSTOM:
         posTable = getPosTable (codePage);
         negTable = getNegTable (codePage);
         zeroPad = EBCDIC_CUST_ZERO_PAD;
         posor =  EBCDIC_CUST_EXTDEC_POSOR;
         negor = EBCDIC_CUST_EXTDEC_NEGOR;
         plus = EBCDIC_CUST_PLUS;
         minus = EBCDIC_CUST_MINUS;
         break;

      case MarshallExternalDecimalUtils.EXTERNAL_DECIMAL_SIGN_ASCII:
         zeroPad = ASCII_ZERO_PAD;
         posor = ASCII_EXTDEC_POSOR;
         negor = ASCII_EXTDEC_NEGOR;
         plus = ASCII_PLUS;
         minus = ASCII_MINUS;
         break;

      default:
      throw new IllegalArgumentException(MarshallResource.instance().getString(MarshallResource.MARSHRT_EXTDEC_SIGN_NOT_VALID, Integer.toString(externalDecimalSign)));
   }

   int i = size - 1;

   if (signed && (signFormat == MarshallExternalDecimalUtils.SIGN_FORMAT_TRAILING_SEPARATE))    // Leave last byte for sign
      i = size - 2;

   // Fill the byte array with the numerical values or'd with 0xF0 (mvs) or 0x30 (ascii)
   for (; (i >= 0 && (binary != 0)); i--)
   {
      // Get the rightmost digit
      int decimalDigit = (int) Math.abs (binary % 10);
      binary = (long) (binary / 10);
      retVal[i] = (byte)(decimalDigit | zeroPad);
   }

   if (i >= 0)    // If the value is less than the size of the field
   {
      // Pad out start of number
      for (int j = i; j >=0; j--)
         retVal[j] = zeroPad;
   }

   if (signed)    // Determine where the sign should be in the buffer
   {
      switch (signFormat)
      {
         case MarshallExternalDecimalUtils.SIGN_FORMAT_LEADING:
            if (externalDecimalSign != MarshallExternalDecimalUtils.EXTERNAL_DECIMAL_SIGN_EBCDIC_CUSTOM)
            {
               if (isPositive)
                  retVal[0] = (byte)((retVal[0] & 0x0F) | posor);
               else
                  retVal[0] = (byte)((retVal[0] & 0x0F) | negor);
            }
            else
            {
               int index = retVal[0] & 0x0F; // Get the value of the digit and look up the value in the table
               if (isPositive)
                  retVal[0] = posTable[index];
               else
                  retVal[0] = negTable[index];
            }

            break;

         case MarshallExternalDecimalUtils.SIGN_FORMAT_LEADING_SEPARATE:
            if (isPositive)
               retVal[0] = plus;
            else
               retVal[0] = minus;

            break;

         case MarshallExternalDecimalUtils.SIGN_FORMAT_TRAILING:
            if (externalDecimalSign != MarshallExternalDecimalUtils.EXTERNAL_DECIMAL_SIGN_EBCDIC_CUSTOM)
            {
               if (isPositive)
                  retVal[size-1] = (byte)((retVal[size-1] & 0x0F) | posor);
               else
                  retVal[size-1] = (byte)((retVal[size-1] & 0x0F) | negor);
            }
            else
            {
               int index = retVal[size-1] & 0x0F;  // Get the value of the digit and look up the value in the table
               if (isPositive)
                  retVal[size-1] = posTable[index];
               else
                  retVal[size-1] = negTable[index];
            }

            break;

         case MarshallExternalDecimalUtils.SIGN_FORMAT_TRAILING_SEPARATE:
            if (isPositive)
               retVal[size-1] = plus;
            else
               retVal[size-1] = minus;

            break;

         default:
         throw new IllegalArgumentException(MarshallResource.instance().getString(MarshallResource.MARSHRT_EXTDEC_SIGN_FMT_NOT_VALID, Integer.toString(signFormat)));
      }
   }

   return (retVal);
}


/**
 * Transforms bidirectional text before data marshalling.
 * @return Transformed bidirectional text.
 * @param value String to transform.
 * @param textType The type of text.  Valid values for this parameter are
 * <code>BidiFlag.TYPE_IMPLICIT</code>, and <code>BidiFlag.TYPE_VISUAL</code>.
 * @param orientation The orientation of the text. Valid values for this parameter are
 * the following:
 * <ul>
 * <li><code>BidiFlag.ORIENTATION_LTR</code></li>
 * <li><code>BidiFlag.ORIENTATION_RTL</code></li>
 * <li><code>BidiFlag.ORIENTATION_CONTEXT_LTR</code></li>
 * <li><code>BidiFlag.ORIENTATION_CONTEXT_RTL</code></li>
 * </ul>
 * @param swap The symmetric swapping of the text. Valid values for this parameter are
 * the following:
 * <ul>
 * <li><code>BidiFlag.SWAP_YES</code></li>
 * <li><code>BidiFlag.SWAP_NO</code></li>
 * </ul>
 * @param numeralShapes The numeral shapes.  Valid values for this parameter are the
 * following:
 * <ul>
 * <li><code>BidiFlag.NUMERALS_NOMINAL</code></li>
 * <li><code>BidiFlag.NUMERALS_NATIONAL</code></li>
 * <li><code>BidiFlag.NUMERALS_CONTEXTUAL</code></li>
 * </ul>
 * @param textShape The text shapes.  Valid values for this parameter are the following:
 * <ul>
 * <li><code>BidiFlag.TEXT_NOMINAL</code></li>
 * <li><code>BidiFlag.TEXT_SHAPED</code></li>
 * <li><code>BidiFlag.TEXT_INITIAL</code></li>
 * <li><code>BidiFlag.TEXT_MIDDLE</code></li>
 * <li><code>BidiFlag.TEXT_FINAL</code></li>
 * <li><code>BidiFlag.TEXT_ISOLATED</code></li>
 * </ul>
 */

public static String transformBidiTextMarshall (String value, BidiFlag textType, BidiFlag orientation,
      BidiFlag swap, BidiFlag numeralShapes, BidiFlag textShape) throws IllegalArgumentException
{
   BidiFlagSet defaultFlagSet = new BidiFlagSet(BidiFlag.TYPE_IMPLICIT, BidiFlag.ORIENTATION_LTR,
      BidiFlag.SWAP_YES, BidiFlag.NUMERALS_NOMINAL, BidiFlag.TEXT_NOMINAL);

   BidiText src = new BidiText(defaultFlagSet, value);

   BidiFlagSet specifiedFlagSet = new BidiFlagSet(textType, orientation, swap, numeralShapes, textShape);

   BidiText dest = src.transform (specifiedFlagSet);

   return (dest.toString());
}

/**
 * Transforms bidirectional text after data marshalling.
 * @return Transformed bidirectional text.
 * @param value String to transform.
 * @param textType The type of text.  Valid values for this parameter are
 * <code>BidiFlag.TYPE_IMPLICIT</code>, and <code>BidiFlag.TYPE_VISUAL</code>.
 * @param orientation The orientation of the text. Valid values for this parameter are
 * the following:
 * <ul>
 * <li><code>BidiFlag.ORIENTATION_LTR</code></li>
 * <li><code>BidiFlag.ORIENTATION_RTL</code></li>
 * <li><code>BidiFlag.ORIENTATION_CONTEXT_LTR</code></li>
 * <li><code>BidiFlag.ORIENTATION_CONTEXT_RTL</code></li>
 * </ul>
 * @param swap The symmetric swapping of the text. It is set to <code>true</code> if swapping
 * is allowed; <code>false</code> otherwise.
 * @param numeralShapes The numeral shapes.  Valid values for this parameter are the
 * following:
 * <ul>
 * <li><code>BidiFlag.NUMERALS_NOMINAL</code></li>
 * <li><code>BidiFlag.NUMERALS_NATIONAL</code></li>
 * <li><code>BidiFlag.NUMERALS_CONTEXTUAL</code></li>
 * </ul>
 * @param textShape The text shapes.  Valid values for this parameter are the following:
 * <ul>
 * <li><code>BidiFlag.TEXT_NOMINAL</code></li>
 * <li><code>BidiFlag.TEXT_SHAPED</code></li>
 * <li><code>BidiFlag.TEXT_INITIAL</code></li>
 * <li><code>BidiFlag.TEXT_MIDDLE</code></li>
 * <li><code>BidiFlag.TEXT_FINAL</code></li>
 * <li><code>BidiFlag.TEXT_ISOLATED</code></li>
 * </ul>
 */

public static String transformBidiTextUnmarshall (String value, BidiFlag textType, BidiFlag orientation,
      BidiFlag swap, BidiFlag numeralShapes, BidiFlag textShape) throws IllegalArgumentException
{
   BidiFlagSet specifiedFlagSet = new BidiFlagSet(textType, orientation, swap, numeralShapes, textShape);

   BidiText src = new BidiText(specifiedFlagSet, value);

   BidiFlagSet defaultFlagSet = new BidiFlagSet(BidiFlag.TYPE_IMPLICIT, BidiFlag.ORIENTATION_LTR,
      BidiFlag.SWAP_YES, BidiFlag.NUMERALS_NOMINAL, BidiFlag.TEXT_NOMINAL);

   BidiText dest = src.transform (defaultFlagSet);

   return (dest.toString());
}

public static void dumpBytes (StringBuffer stream, byte[] buffer)
{
   int bufSize = 0;

   if (buffer != null)
      bufSize = buffer.length;
   int size = bufSize;
   char buf[] = new char[16];

   while ((size % 16) != 0)
      size++;

   // Output record size
   stream.append("Buffer Size: "+bufSize+" bytes");
   stream.append("\n");
   stream.append("\n");

   if (buffer == null)
      return;

   int j = 0;

   for (int i = 1; i <= size; i++)
   {
      if (i > bufSize)
      {
         stream.append(Integer.toHexString(0));
         stream.append(Integer.toHexString(0));
         buf[j] = 0;
      }
      else
      {
         int k = (int)((buffer[i-1] & 0xF0) >> 4);
         int l = (int)(buffer[i-1] & 0x0F);
         stream.append(Integer.toHexString(k));
         stream.append(Integer.toHexString(l));

         buf[j] = (char)buffer[i-1];
      }

      // Group hex dump into 4 byte chunks
      if ((i % 4) == 0)
         stream.append (" ");

      // Display bytes as printable characters
      if ((i % 16) == 0)
      {
         stream.append(" |");
         for (int k = 0; k < 16; k++)
         {
            if (Character.isIdentifierIgnorable(buf[k]))
               stream.append(".");
            else
               stream.append(buf[k]);
         }

         stream.append("|");
         stream.append("\n");
         j = 0;
      }
      else
         j++;
   }

   stream.append("\n");

   return;
}



/*
 * Only the first 80 bits (10 bytes) of storage are used in the extended Intel format,
 * but it will occupy 16 bytes of storage in a structure.  Therfore we can ignore the
 * last 6 bytes.
 *
 * Bit mapping:
 * In the internal representation, there is 1 bit for the sign (S), 15 bits for the
 * exponent (E), and 64 bits for the fraction (F).
 *    _ ____________________ ___________/___________ _____/______
 *   | |  Exponent + 16383  |       Fraction        |            |
 *   |S|     (15 bits)      |      (64 bits)        |  Unused    |
 *   |_|____________________|___________/___________|_____/______|
 *    0 1                   16                     79          127
 * The input byte array is expected to be in Big Endian format
 * 
 * There are really only 63 bits of precision, as the top bit of the fraction
 * is set explicitly to 0 or 1 as outlined below.
 * 
 * For the single-precision and double-precision formats, only the fraction part of the
 * significand is encoded.  The integer is assumed to be 1 for all numbers except 0 and
 * denormalized finite numbers.  For the double extended-precision format, the integer
 * is contained in bit 63, and the most-significant fraction bit is bit 62.  Here, the
 * integer is explicitly set to 1 for normalized numbers, infinites, and NaNs, and to 0
 * for zero and denormalized numbers.
 */

public static String convertIEEEExtendedIntelToString (byte[] bytes)
{
   // The first bit is the sign bit

   boolean isPositive = true;
   if ((bytes[0] & BYTE_SIGN_MASK) != 0)
      isPositive = false;

   // The bits 1 to 15 (15 bits) contain the biased exponent (exponent + bias)

   byte[] exponentBytes = new byte[2];
   System.arraycopy (bytes, 0, exponentBytes, 0, 2);
   // Mask out the sign bit
   exponentBytes[0] = (byte)(exponentBytes[0] & (byte)0x7F);

   // Convert the exponent bytes into a number
   short exponent = ConversionUtils.convertFromByteToShort (exponentBytes);

   // Subtract the exponent bias from the exponent to get the real exponent
   short realExponent = -16382;     // Minimum exponent
   if (exponent != 0)
      realExponent = (short)(exponent - 16383);

   // Bits 16 to 80 contain the fraction or mantissa (64 bits)
   byte[] fractionBytes = new byte[9];
   System.arraycopy (bytes, 2, fractionBytes, 1, 8);

   BigInteger fraction = new BigInteger(fractionBytes);

   // Handle +- 0 and +- Infinity
   if (exponent != 0)
   {
      if (exponent == 0x7FFF) // +- Infinity or NaN
      {
         if (!fraction.equals(BigInteger.ZERO)) // NaN
            return ("NaN");
         else if (isPositive)
            return ("Infinity");
         else
            return ("-Infinity");
      }
   }
   else  // Unnormalized number or +- 0 if (fraction == 0)
   {
      if (fraction.equals(BigInteger.ZERO))
      {
         if (isPositive)
            return ("0");
         else
            return ("-0");
      }
   }

   // Need to shift the fraction bits to the right up to the first bit set.
   // For each shift right (multiply of 2) we need to reduce the exponent by one (divide by 2).
   int firstNonZero = fraction.getLowestSetBit();

   int numShift = 63;      // Number of fraction bits
   if (exponent == 0)      // Denormal number
   {
      // V = (-1)**S x 2**(-16382) x (0.F)
      // Do not need to shift the fraction bits as there is no implicit one, but
      // need to reduce the exponent by a value of the lowest bit set.
      numShift = numShift - firstNonZero;       // Number of bits shifting number to the left
      realExponent -= numShift;
   }
   else
   {
      if (firstNonZero != 0)
      {
         firstNonZero--;
         numShift = numShift - firstNonZero;    // Number of bits really shifting number to the left
         fraction = fraction.shiftRight (firstNonZero);
      }
      realExponent -= numShift;     // Reduce the exponent by the number of shifts
   }

   String result = getFractionAsString (isPositive, realExponent, fraction, 2, 20);

   return (result);
}



public static byte[] convertBigDecimalToIEEEExtendedIntel (BigDecimal fpNumber)
{
   byte[] result = new byte[10];

   // Handle the special case of zero immediately
   if (fpNumber.signum() == 0)
      return (result);

   // The first bit is the sign bit
   boolean isPositive = true;

   if (fpNumber.signum() == -1)
   {
      isPositive = false;
      result[0] = (byte)(result[0] | BYTE_SIGN_MASK); // Set the sign bit
   }

   fpNumber = fpNumber.abs();    // No longer need sign
   BigInteger[] base2 = getBase2FractionAndExponent (fpNumber, 63, true);
   BigInteger realFraction = base2[0];

   // If the exponent bias - abs(exponent) < 0, we have a denormal number.
   // Set the exponent to zero and shift the fraction over.
   // Exponent bias for an IEEE Extended Intel is 16383.
   short realExponent = base2[1].shortValue();
   short biasedExponent = 16383;
   int numShifts = biasedExponent - Math.abs(realExponent);

   if (numShifts < 0)   // Denormal - fix fraction and exponent
   {
      realFraction = realFraction.setBit (63);  // Put back the implicit one bit
      realFraction = realFraction.shiftRight (Math.abs(numShifts)+1);
      realFraction = realFraction.clearBit (63);  // Clear the top bit to denote denormal
      biasedExponent = 0;
   }
   else
   {
      biasedExponent += realExponent;
      realFraction = realFraction.setBit (63);  // Set the top bit to denote normalized
   }

   byte[] fractionBytes = realFraction.toByteArray();

   // If the number is denormal, we won't get back 8 bytes from the above call,
   // so we need to put them into the correct spot in the byte array.
   // Also there could be an empty byte at beginning from the toByteArray() call
   // as an inital zero may have been added to ensure the number is positive.

   if (fractionBytes.length > 8)
   {
      byte[] newBytes = new byte[8];
      int i = fractionBytes.length - 8;
      System.arraycopy (fractionBytes, i, newBytes, 0, 8);
      fractionBytes = newBytes;
   }

   int offset = 2;
   if (numShifts < 0)
      offset = 10 - fractionBytes.length;

   // Bits 16 to 80 contain the fraction or mantissa (63+1 bits)
   System.arraycopy (fractionBytes, 0, result, offset, Math.min (fractionBytes.length,8));

   // The bits 1 to 15 (15 bits) contain the biased exponent (exponent + bias)

   byte[] exponentBytes = ConversionUtils.convertToByte (biasedExponent);
   result[0] = (byte)(exponentBytes[0] | result[0]);
   result[1] = exponentBytes[1];

   return (result);
}

/*
 * The format of an OS390 IEEE Extended FP number is as follows:
 *    _ ____________________ ___________/___________
 *   | |  Exponent + 16383  |       Fraction        |
 *   |S|     (15 bits)      |      (112 bits)       |
 *   |_|____________________|___________/___________|
 *    0 1                   16                    127
 *
 * The input byte array is expected to be in Big Endian format
 */

public static String convertIEEEExtendedOS390ToString (byte[] bytes)
{
   // The first bit is the sign bit

   boolean isPositive = true;
   if ((bytes[0] & BYTE_SIGN_MASK) != 0)
      isPositive = false;

   // The bits 1 to 15 (15 bits) contain the biased exponent (exponent + bias)

   byte[] exponentBytes = new byte[2];
   System.arraycopy (bytes, 0, exponentBytes, 0, 2);
   // Mask out the sign bit
   exponentBytes[0] = (byte)(exponentBytes[0] & (byte)0x7F);

   // Convert the exponent bytes into a number
   short exponent = ConversionUtils.convertFromByteToShort (exponentBytes);

   // Subtract the exponent bias from the exponent to get the real exponent
   short realExponent = -16382;     // Minimum exponent
   if (exponent != 0)
      realExponent = (short)(exponent - 16383);

   // Bits 16 to 127 contain the fraction or mantissa (112 bits)
   byte[] fractionBytes = new byte[15];
   System.arraycopy (bytes, 2, fractionBytes, 1, 14);

   BigInteger fraction = new BigInteger(fractionBytes);

   // Add the implicit one if exponent is not zero, and handle +- 0 and +- Infinity
   if (exponent != 0)
   {
      if (exponent == 0x7FFF) // +- Infinity or NaN
      {
         if (!fraction.equals(BigInteger.ZERO)) // NaN
            return ("NaN");
         else if (isPositive)
            return ("Infinity");
         else
            return ("-Infinity");
      }

      // Add the implicit one before the decimal point
      fractionBytes[0] = (byte)(fractionBytes[0] | 0x01);
      fraction = new BigInteger(fractionBytes);
   }
   else  // Unnormalized number or +- 0 if (fraction == 0)
   {
      if (fraction.equals(BigInteger.ZERO))
      {
         if (isPositive)
            return ("0");
         else
            return ("-0");
      }
   }

   // Need to shift the fraction bits to the right up to the first bit set.
   // For each shift right (multiply of 2) we need to reduce the exponent by one (divide by 2).
   int firstNonZero = fraction.getLowestSetBit();

   int numShift = 112;        // Number of fraction bits
   if (exponent == 0)      // Denormal number
   {
      // V = (-1)**S x 2**(-16382) x (0.F)
      // Do not need to shift the fraction bits as there is no implicit one, but
      // need to reduce the exponent by a value of the lowest bit set.
      numShift = numShift - firstNonZero;       // Number of bits shifting number to the left
      realExponent -= numShift;
   }
   else
   {
      if (firstNonZero != 0)
      {
         firstNonZero--;
         numShift = 112 - firstNonZero;      // Number of bits really shifting number to the left
         fraction = fraction.shiftRight (firstNonZero);
      }
      realExponent -= numShift;     // Reduce the exponent by the number of shifts
   }

   String result = getFractionAsString (isPositive, realExponent, fraction, 2, 34);

   return (result);
}

public static byte[] convertBigDecimalToIEEEExtendedOS390 (BigDecimal fpNumber)
{
   byte[] result = new byte[16];

   // Handle the special case of zero immediately
   if (fpNumber.signum() == 0)
      return (result);

   // The first bit is the sign bit
   boolean isPositive = true;

   if (fpNumber.signum() == -1)
   {
      isPositive = false;
      result[0] = (byte)(result[0] | BYTE_SIGN_MASK); // Set the sign bit
   }

   fpNumber = fpNumber.abs();    // No longer need sign
   BigInteger[] base2 = getBase2FractionAndExponent (fpNumber, 112, true);
   BigInteger realFraction = base2[0];

   // If the exponent bias - abs(exponent) < 0, we have a denormal number.
   // Set the exponent to zero and shift the fraction over.
   // Exponent bias for an IEEE Extended OS390 is 16383.
   short realExponent = base2[1].shortValue();
   short biasedExponent = 16383;
   int numShifts = biasedExponent - Math.abs(realExponent);

   if (numShifts < 0)   // Denormal - fix fraction and exponent
   {
      realFraction = realFraction.setBit (112); // Put back the implicit one bit
      realFraction = realFraction.shiftRight (Math.abs(numShifts)+1);
      biasedExponent = 0;
   }
   else
      biasedExponent += realExponent;

   byte[] fractionBytes = realFraction.toByteArray();

   // If the number is denormal, we won't get back 8 bytes from the above call,
   // so we need to put them into the correct spot in the byte array.
   // Also there could be an empty byte at beginning from the toByteArray() call
   // as an inital zero may have been added to ensure the number is positive.

   if (fractionBytes.length > 14)
   {
      byte[] newBytes = new byte[14];
      int i = fractionBytes.length - 14;
      System.arraycopy (fractionBytes, i, newBytes, 0, 14);
      fractionBytes = newBytes;
   }

   int offset = 2;
   if (numShifts < 0)
      offset = 16 - fractionBytes.length;

   // Bits 16 to 127 contain the fraction or mantissa (112 bits)
   System.arraycopy (fractionBytes, 0, result, offset, Math.min (fractionBytes.length,14));

   // The bits 1 to 15 (15 bits) contain the biased exponent (exponent + bias)

   byte[] exponentBytes = ConversionUtils.convertToByte (biasedExponent);
   result[0] = (byte)(exponentBytes[0] | result[0]);
   result[1] = exponentBytes[1];

   return (result);
}


/*
 * Bit mapping:
 * A 128-bit long double number consists of an ordered pair of 64-bit double-precision numbers.
 * The first member of the ordered pair contains the high-order part of the number, and the second
 * member contains the low-order part. The value of the long double quantity is the sum of the two
 * 64-bit numbers.
 *
 * Each of the two 64-bit numbers is itself a double-precision floating-point number with a sign,
 * exponent, and significand. Typically the low-order member has a magnitude that is less than 0.5 units
 * in the last place of the high part, so the values of the two 64-bit numbers do not overlap and the
 * entire significand of the low-order number adds precision beyond the high-order number.
 *
 * The input byte array is expected to be in Big Endian format
 */

public static String convertIEEEExtendedAIXToString (byte[] bytes)
{
   byte[] double1Bytes = new byte[8];
   byte[] double2Bytes = new byte[8];

   System.arraycopy (bytes, 0, double1Bytes, 0, 8);
   System.arraycopy (bytes, 8, double2Bytes, 0, 8);

   double first = ConversionUtils.convertFromByteToDouble (double1Bytes);
   double second = ConversionUtils.convertFromByteToDouble (double2Bytes);

   BigDecimal bd1 = new BigDecimal (Double.toString(first));
   BigDecimal bd2 = new BigDecimal (Double.toString(second));

   BigDecimal res = bd1.add (bd2);

   String result = res.toString();

   return (result);
}

public static byte[] convertBigDecimalToIEEEExtendedAIX (BigDecimal fpNumber)
{
   byte[] result = new byte[16];

   // Handle the special case of zero immediately
   if (fpNumber.signum() == 0)
      return (result);

   // The magnitude of the number will not exceed the magnitude of a double,
   // but the precision can be greater.  Convert the number to a double and
   // then subtract it from itself to see if there is any remainder.

   double num = fpNumber.doubleValue();

   if (num == Double.POSITIVE_INFINITY)
      num = Double.MAX_VALUE;
   else if (num == Double.NEGATIVE_INFINITY)
      num = -Double.MAX_VALUE;

   byte[] numBytes = convertToByte (num);

   // Copy the high order part to the result
   System.arraycopy (numBytes, 0, result, 0, numBytes.length);

   BigDecimal bdNum = new BigDecimal(Double.toString(num));

   BigDecimal res = fpNumber.subtract (bdNum);

   // If there is a remainder convert it and copy it to the low order part
   if (res.signum() != 0)
   {
      num = res.doubleValue();
      numBytes = convertToByte (num);
      System.arraycopy (numBytes, 0, result, 8, numBytes.length);
   }

   return (result);
}


/*
 * Converts a BigInteger fraction and exponent to a base 10 number string for a
 * given exponent base.  The returned string length is up to the given number of
 * significant digits.
 */

private static String getFractionAsString (boolean isPositive, short realExponent, BigInteger fraction, int exponentBase, int numSignificantDigits)
{
   String result;
   BigInteger biExponent = new BigInteger(Integer.toString(exponentBase));
   biExponent = biExponent.pow (Math.abs(realExponent));

   BigInteger biResult;

   if (realExponent < 0)      // Divide fraction by exponent
   {
      biResult = fraction.divide (biExponent);
      StringBuffer sb = new StringBuffer();
      if (biResult.equals(BigInteger.ZERO))
         sb.append ("0");
      else
         sb.append(biResult.toString());

      int size = sb.length();

      BigInteger remainder = fraction.remainder (biExponent);

      if (!remainder.equals(BigInteger.ZERO))
         sb.append (".");

      BigInteger ten = new BigInteger("10");

      boolean foundNonZero = false;

      while ((!remainder.equals(BigInteger.ZERO)) && (size <= numSignificantDigits))
      {
         // Need to process remainder up to the precision of the floating point number
         remainder = remainder.multiply (ten);
         BigInteger divResult;
         if (size != numSignificantDigits)
            divResult = remainder.divide (biExponent);
         else     // Round the last significant digit
            divResult = ((new BigDecimal(remainder)).divide(new BigDecimal(biExponent), BigDecimal.ROUND_HALF_UP)).toBigInteger();

         sb.append (divResult.toString());
         if (!foundNonZero)
         {
            if (!divResult.equals(BigInteger.ZERO))      // Only start counting significant digits after
               foundNonZero = true;                   // the first non-zero number.
         }
         else
            size++;
         remainder = remainder.remainder (biExponent);
      }

      result = sb.toString();
   }
   else     // Multiply fraction by exponent
   {
      biResult = fraction.multiply (biExponent);
      result = biResult.toString();
   }

   if (!isPositive)
      result = "-" + result;

   return result;
}

/*
 * Returns the base 2 fraction and exponent for a BigDecimal base 10 number.
 * @return Array of two BigIntegers, the first one being the fraction and the second
 * one being the exponent.
 * @param fpNumber The floating point number.
 * @param numFractionBits The maximun number of fraction bits for the given floating point number.
 * @param addPrecisonForImplicitOne Flag to indicate to add an extra bit of precision
 */

private static BigInteger[] getBase2FractionAndExponent (BigDecimal fpNumber, int numFractionBits, boolean addPrecisonForImplicitOne)
{
   BigInteger integralPart = fpNumber.toBigInteger();

   fpNumber = fpNumber.subtract(new BigDecimal(integralPart));

   BigInteger fractionalPart = BigInteger.ZERO;
   int exponent = 0;

   if (fpNumber.signum() != 0)      // There is a fractional part to the number
   {
//    // Mulitply the fraction by 2 ** exponent which is greater than 10 ** scale
//    double lntenpow = StrictMath.log (StrictMath.pow (10D, -fpNumber.scale()));
//    double res = lntenpow / StrictMath.log(2);
//    BigDecimal exp = new BigDecimal (res);
//    exp = exp.divide (new BigDecimal(BigInteger.ONE), 0,BigDecimal.ROUND_UP);
//    exp = exp.abs();
//
//    int maxExponent = 0;
//    for (int i = 0; i < 15; i++)
//    {
//       maxExponent <<= 1;
//       maxExponent |= 0x01;
//    }
//    BigInteger two = new BigInteger("2");
//    fpNumber = fpNumber.multiply(new BigDecimal(two.pow(exp.intValue())));
//
//    fractionalPart = fpNumber.toBigInteger();
//    exponent = -exp.intValue();

      // For every multiply by two, it is really shifting left by one bit.
      BigDecimal two = new BigDecimal("2");
      BigDecimal one = new BigDecimal("1");
      boolean firstSetBitFound = false;
      int i = 0;
      while ((i <= numFractionBits) && (fpNumber.signum() != 0))
      {
         fpNumber = fpNumber.multiply (two);
         exponent--;
         fractionalPart = fractionalPart.shiftLeft (1);
         if (fpNumber.toBigInteger().compareTo(BigInteger.ONE) == 0)
         {
            firstSetBitFound = true;
            fractionalPart = fractionalPart.setBit(0);
            fpNumber = fpNumber.subtract (one);
         }

         if (firstSetBitFound)
         {
            if (i == numFractionBits)
            {
               if (fpNumber.compareTo(new BigDecimal("0.5")) >= 0)   // Round up fractional part
                  fractionalPart = fractionalPart.add (BigInteger.ONE);
            }
            i++;
         }
      }
   }

   // The fractional part value is now fractionalPart * 2 ** exponent

   if (exponent != 0)   // Has a fractional part
   {
      // Shift the integral part to the left by the exponent and or the exponent into it
      integralPart = integralPart.shiftRight (exponent);
      integralPart = integralPart.or (fractionalPart);
   }

   // Now we need to normalize the integral part to put the decimal point to the very left.
   // Number of bits set to the left of the decimal point is the bit length.
   int bitLength = integralPart.bitLength();

   // Need to increase the exponent by the bit length
   exponent += bitLength;

   // Need to shift the integral part over to line it up with the number of fraction bits
   if (bitLength > numFractionBits)
   {
      int rightShift = bitLength - numFractionBits;
      if (addPrecisonForImplicitOne)
         rightShift--;
      integralPart = integralPart.shiftRight (rightShift);
   }
   else
   {
      int leftShift = numFractionBits - bitLength;
      if (addPrecisonForImplicitOne)
         leftShift++;
      integralPart = integralPart.shiftLeft (leftShift);
   }

   if (addPrecisonForImplicitOne)   // Clear the high bit and reduce the exponent
   {
      integralPart = integralPart.clearBit (integralPart.bitLength() - 1);
      exponent--;       // Reduce the exponent by one for the left shift
   }

   BigInteger[] result = new BigInteger[2];
   result[0] = integralPart;
   result[1] = new BigInteger(Integer.toString(exponent));

   return (result);
}


/*
 *
private static BigInteger[] getBase2FractionAndExponent (BigDecimal fpNumber, int numFractionBits, boolean addPrecisonForImplicitOne)
{
   BigDecimal one = new BigDecimal("1");

   // Number of bits set to the left of the decimal point is the exponent if the value is > 1
   int exponent = fpNumber.toBigInteger().bitLength();
   // Number of significant bits is numFractionBits - exponent (# bits set to left of decimal point)
   int sigBits = numFractionBits - exponent;

   boolean lessThanOne = false;
   if (fpNumber.compareTo(one) < 0) // Less than one
      lessThanOne = true;

   BigInteger realFraction = null;

   if (sigBits < 0)     // A really big exponent
   {
      realFraction = fpNumber.toBigInteger();
      realFraction = realFraction.shiftRight (exponent - numFractionBits - 1);
   }
   else
   {
      // Shift left the value by the number of sigBits to get all of the fraction bits
      // to the left of the decmal point (multiply by 2 ** sigBits)
      BigInteger two = new BigInteger("2");
      BigDecimal fraction = fpNumber.multiply(new BigDecimal(two.pow(sigBits)));
      int bitLength = fraction.toBigInteger().bitLength();

      // If the value is less than one, then need to find 1st bit set after the decimal
      // point for the exponent (it will be a negative exponent)
      if (lessThanOne)
         exponent = bitLength - numFractionBits;

      // Need to figure out where the first 'one' or set bit is in the fraction,
      // so we can shift the fraction to the left to take care of the implicit one.
      // bitLength() returns the number of bits in the ordinary binary representation,
      // so the first bit set will be equal to this number.
      // Shift the value to the left by (numFractionBits - bitLength + 1) for the implicit one.
      // This puts the first set bit at position numFractionBits+1, allowing the greatest precision.
      int leftShift = numFractionBits - bitLength;
      if (addPrecisonForImplicitOne)
      {
         leftShift++;      // Shift one more left for implicit one
         exponent--;       // Reduce the exponent by one for the left shift
      }

      // Now multiply the initial number by leftShift+sigDigits to get the normalized fraction
      fraction = fpNumber.multiply(new BigDecimal(two.pow(sigBits+leftShift)));

      // Round up the last significant bit
      fraction = fraction.divide (one, 0, BigDecimal.ROUND_HALF_UP);
      realFraction = fraction.toBigInteger();
   }

   // Clear the first bit set in the fraction to remove the implicit one
   if (addPrecisonForImplicitOne)
      realFraction = realFraction.clearBit (realFraction.bitLength() - 1);

   BigInteger[] result = new BigInteger[2];
   result[0] = realFraction;
   result[1] = new BigInteger(Integer.toString(exponent));

   return (result);
}
 *
 */

/*
 *
 * IEEE 754 Floating point format represents double precision numbers as follows:
 *
 * In the internal representation, there is 1 bit for the sign (S), 11 bits for the
 * exponent (E), and 52 bits for the fraction (F).
 *    _ ____________________ ___________/___________
 *   | |  Exponent + 1023   |       Fraction        |
 *   |S|     (11 bits)      |      (52 bits)        |
 *   |_|____________________|___________/___________|
 *    0 1                   12                     64
 *
 * The input byte array is expected to be in Big Endian format.
 */

public static String convertIEEEDoubleToString (byte[] bytes)
{
   // The first bit is the sign bit

   boolean isPositive = true;
   if ((bytes[0] & BYTE_SIGN_MASK) != 0)
      isPositive = false;

   // The bits 1 to 11 (11 bits) contain the biased exponent (exponent + bias)

   byte[] exponentBytes = new byte[2];
   System.arraycopy (bytes, 0, exponentBytes, 0, 2);
   // Mask out the sign bit
   exponentBytes[0] = (byte)(exponentBytes[0] & (byte)0x7F);
   // Mask out bits 12 to 16
   exponentBytes[1] = (byte)(exponentBytes[1] & (byte)0xF0);

   // Convert the exponent bytes into a number
   short exponent = ConversionUtils.convertFromByteToShort (exponentBytes);
   // Shift it over 4 bits
   exponent = (short)(exponent >> 4);

   // Subtract the exponent bias from the exponent to get the real exponent
   short realExponent = -1022;      // Minimum exponent
   if (exponent != 0)
      realExponent = (short)(exponent - 1023);

   // Bits 12 to 63 contain the fraction or mantissa (52 bits)
   byte[] fractionBytes = new byte[8];
   System.arraycopy (bytes, 0, fractionBytes, 0, 8);
   fractionBytes[0] = 0;
   fractionBytes[1] = (byte)(fractionBytes[1] & 0x0F);

   BigInteger fraction = new BigInteger(fractionBytes);

   // Add the implicit one if exponent is not zero, and handle +- 0 and +- Infinity
   if (exponent != 0)
   {
      if (exponent == 0x7FF)  // +- Infinity or NaN
      {
         if (!fraction.equals(BigInteger.ZERO)) // NaN
            return ("NaN");
         else if (isPositive)
            return ("Infinity");
         else
            return ("-Infinity");
      }

      // Add the implicit one before the decimal point
      fractionBytes[1] = (byte)(fractionBytes[1] | 0x10);
      fraction = new BigInteger(fractionBytes);
   }
   else  // Denormalized number or +- 0 if (fraction == 0)
   {
      if (fraction.equals(BigInteger.ZERO))
      {
         if (isPositive)
            return ("0");
         else
            return ("-0");
      }
   }


   int firstNonZero = fraction.getLowestSetBit();

   int numShift = 52;      // Number of fraction bits
   if (exponent == 0)      // Denormal number
   {
      // V = (-1)**S x 2**(-1022) x (0.F)
      // Do not need to shift the fraction bits as there is no implicit one, but
      // need to reduce the exponent by a value of the lowest bit set.
      numShift = numShift - firstNonZero;       // Number of bits shifting number to the left
      realExponent -= numShift;
   }
   else
   {
      // Need to shift the fraction bits to the right up to the first bit set.
      // For each shift right (multiply of 2) we need to reduce the exponent by one (divide by 2).
      if (firstNonZero != 0)
      {
         firstNonZero--;
         numShift = numShift - firstNonZero;       // Number of bits really shifting number to the left
         fraction = fraction.shiftRight (firstNonZero);
      }
      realExponent -= numShift;     // Reduce the exponent by the number of shifts
   }

   String result = getFractionAsString (isPositive, realExponent, fraction, 2, 16);

   return (result);
}


public static byte[] convertBigDecimalToDouble (BigDecimal fpNumber)
{
   byte[] result = new byte[8];

   // Handle the special case of zero immediately
   if (fpNumber.signum() == 0)
      return (result);

   // The first bit is the sign bit
   boolean isPositive = true;

   if (fpNumber.signum() == -1)
   {
      isPositive = false;
      result[0] = (byte)(result[0] | BYTE_SIGN_MASK); // Set the sign bit
   }

   fpNumber = fpNumber.abs();    // No longer need sign
   BigInteger[] base2 = getBase2FractionAndExponent (fpNumber, 52, true);
   BigInteger realFraction = base2[0];

   // If the exponent bias - abs(exponent) < 0, we have a denormal number.
   // Set the exponent to zero and shift the fraction over.
   // Exponent bias for a double is 1023
   short realExponent = base2[1].shortValue();
   short biasedExponent = 1023;
   int numShifts = biasedExponent - Math.abs(realExponent);

   if (numShifts < 0)   // Denormal - fix fraction and exponent
   {
      realFraction = realFraction.setBit (52);  // Put back the implicit one bit
      realFraction = realFraction.shiftRight (Math.abs(numShifts)+1);
      biasedExponent = 0;
   }
   else
      biasedExponent += realExponent;

   byte[] fractionBytes = realFraction.toByteArray();

   // If the number is denormal, we won't get back 7 bytes from the above call,
   // so we need to put them into the correct spot in the byte array.
   // Also there could be an empty byte at beginning from the toByteArray() call
   // as an inital zero may have been added to ensure the number is positive.

   if (fractionBytes.length > 7)
   {
      byte[] newBytes = new byte[7];
      int i = fractionBytes.length - 7;
      System.arraycopy (fractionBytes, i, newBytes, 0, 7);
      fractionBytes = newBytes;
   }

   int offset = 1;
   if (numShifts < 0)
      offset = 8 - fractionBytes.length;

   // Bits 12 to 63 contain the fraction or mantissa (52 bits)
   System.arraycopy (fractionBytes, 0, result, offset, Math.min (fractionBytes.length,7));

   // The bits 1 to 15 (15 bits) contain the biased exponent (exponent + bias)
   // Shift the biased exponent 4 bits to the left to align it correctly
   biasedExponent = (short)(biasedExponent << 4);

   byte[] exponentBytes = ConversionUtils.convertToByte (biasedExponent);
   result[0] = (byte)(exponentBytes[0] | result[0]);
   result[1] = (byte)(exponentBytes[1] | result[1]);

   return (result);
}



/*
 * A hexadecimal-floating-point (HFP) number consists of a sign  bit, a hexadecimal fraction,
 * and an unsigned seven-bit binary integer called the characteristic. The characteristic
 * represents a signed  exponent  and  is obtained  by  adding  64  to the exponent value
 * (excess-64 notation).  The range of the characteristic is 0 to 127, which corresponds to an
 * exponent range of -64 to +63.  The magnitude of an HFP number is the product of its fraction
 * and the number 16 raised to the power of the exponent that is represented by its characteristic.
 * The number is positive or negative depending on whether the sign bit is zero or one, respectively.
 *
 * The fraction of an HFP number is treated as a hexadecimal number because it is considered to be
 * multiplied by a number which is a power of 16. The name, fraction, indicates that the radix point
 * is assumed to be immediately to the left of the leftmost fraction digit.
 *
 * The first bit (bit 0) is the sign bit (S). The next seven bits are the characteristic.
 *
 * An extended HFP number has a 28-digit fraction and consists of two long HFP numbers that are called
 * the high-order and low-order parts. The high-order part may be any long HFP number. The fraction of the
 * high-order part contains the leftmost 14 hexadecimal digits of the 28-digit fraction. The characteristic
 * and sign of the high-order part are the characteristic and sign of the extended HFP number. If the
 * high-order part is normalized, the extended number is considered normalized. The fraction of the
 * low-order part contains the rightmost 14 digits of the 28-digit fraction. The sign and characteristic
 * of the low-order part of an extended operand are ignored.
 *
 * When a result is generated in the extended format and placed in a register pair, the sign of the
 * low-order part is made the same as that of the high-order part, and, unless the result is a true
 * zero, the low-order characteristic is made 14 less than the high-order characteristic. When the
 * subtraction of 14 would cause the low-order characteristic to become less than zero, the characteristic
 * is made 128 greater than its correct value. (Thus, the subtraction is performed modulo 128.) HFP exponent
 * underflow is indicated only when the high-order characteristic underflows.
 *
 * When an extended result is made a true zero, both the high-order and low-order parts are made a true zero.
 *
 * Extended HFP Number
 *
 *              High-Order Part
 *   _ ______________ ________/____________
 *  | |  High-Order  | Leftmost 14 Digits  |
 *  |S|Characteristic|of 28-Digit Fraction |
 *  |_|______________|________/____________|
 *  0  1              8                   63
 *
 *              Low-Order Part
 *   _ ______________ ________/____________
 *  | |  Low-Order   |Rightmost 14 Digits  |
 *  |S|Characteristic|of 28-Digit Fraction |
 *  |_|______________|________/____________|
 *  64               72                  127
 *
 *
 * The input byte array is expected to be in Big Endian format
 *
 */

public static String convertIBM390HexExtendedToString (byte[] bytes)
{
   // The first bit is the sign bit

   boolean isPositive = true;
   if ((bytes[0] & BYTE_SIGN_MASK) != 0)
      isPositive = false;

   // The bits 1 to 7 (7 bits) contain the characteristic or biased exponent (exponent + bias)
   byte exponent = bytes[0];

   // Mask out the sign bit
   exponent = (byte)(exponent & (byte)0x7F);

   // Subtract 64 from the exponent to get the real exponent (range of 0 to 127)
   byte realExponent = (byte)(exponent - 64);

   // Bits 8 to 63 and 72 to 127 contain the fraction or mantissa (112 bits)
   // Allocate an extra byte so the BigInteger created will always be positive,
   // as we cannot right shift with a zero extension on a BigInteger.
   byte[] fractionBytes = new byte[15];
   System.arraycopy (bytes, 1, fractionBytes, 1, 7);
   System.arraycopy (bytes, 9, fractionBytes, 8, 7);

   BigInteger fraction = new BigInteger(fractionBytes);

   // Handle +- 0 and +- Infinity
   if (exponent != 0)
   {
      if (exponent == 0x7F)   // +- Infinity or NaN
      {
         if (!fraction.equals(BigInteger.ZERO)) // NaN
            return ("NaN");
         else if (isPositive)
            return ("Infinity");
         else
            return ("-Infinity");
      }
   }
   else  // Unnormalized number or +- 0 if (fraction == 0)
   {
      if (fraction.equals(BigInteger.ZERO))
      {
         if (isPositive)
            return ("0");
         else
            return ("-0");
      }
   }

   // Need to shift the fraction bits to the right up to the first bit set.
   // As the exponent is base 16 rather than base 2, we need to shift right
   // in groups of four bits (multiply of 16), to reduce the exponent by one (divide by 16).
   int firstNonZero = fraction.getLowestSetBit();

   int numShift = 112;  // Number of fraction bits
   if (exponent == 0)      // Denormal number
   {
      // Do not need to shift the fraction bits as there is no implicit one, but
      // need to reduce the exponent by a value of the lowest bit set in a group of 4.
      // The first non zero set will always be on a 4 bit boundary.
      numShift = numShift - firstNonZero;       // Number of bits shifting number to the left
      while ((numShift % 4) != 0)
      {
         numShift--;
         fraction = fraction.shiftLeft (1);
      }
      realExponent -= numShift / 4;
   }
   else
   {
      if (firstNonZero != 0)
      {
         firstNonZero--;      // Number of bits need to shift fraction over
         while ((firstNonZero % 4) != 0)
            firstNonZero--;

         numShift = numShift - firstNonZero;    // Must shift in groups of 4 bits to reduce exponent by 1
         fraction = fraction.shiftRight (firstNonZero);
      }
      realExponent -= numShift / 4;    // Reduce the exponent by the number of shifts divided by 4
   }

   String result = getFractionAsString (isPositive, realExponent, fraction, 16, 34);

   return (result);
}


public static byte[] convertBigDecimalToIBM390HexExtended (BigDecimal fpNumber)
{
   byte[] result = new byte[16];

   // Handle the special case of zero immediately
   if (fpNumber.signum() == 0)
      return (result);

   // The first bit is the sign bit
   boolean isPositive = true;

   if (fpNumber.signum() == -1)
   {
      isPositive = false;
      result[0] = (byte)(result[0] | BYTE_SIGN_MASK); // Set the sign bit
   }

   fpNumber = fpNumber.abs();    // No longer need sign
   BigInteger[] base2 = getBase2FractionAndExponent (fpNumber, 112, false);
   BigInteger realFraction = base2[0];

   // The exponent bias is 64 as the real exponent range is 0 to 127.
   // As the exponent is base 16 rather than base 2, we need to shift right
   // in groups of four bits (divide by 16), to reduce the exponent by one.
   // If the exponent bias - abs(exponent) < 0, we have a denormal number.
   // Set the exponent to zero and shift the fraction over.

   short realExponent = base2[1].shortValue();

   // Need to divide the exponent by 4 to get the base 16 exponent.
   // If it does not divide evenly, we need to keep shifting the fraction
   // to the right by one and increasing the exponent by one.

   while ((realExponent % 4) != 0)
   {
      realFraction = realFraction.shiftRight (1);
      realExponent += 1;
   }

   short base16Exponent = (short)(realExponent / 4);

   short biasedExponent = 64;
   int numShifts = biasedExponent - Math.abs(base16Exponent);

   if (numShifts < 0)   // Denormal - fix fraction and exponent
   {
      realFraction = realFraction.shiftRight (Math.abs(numShifts*4));
      biasedExponent = 0;
   }
   else
      biasedExponent += base16Exponent;

   byte[] fractionBytes = realFraction.toByteArray();

   // If the number is denormal, we won't get back 14 bytes from the above call,
   // so we need to put them into the correct spot in the byte array.
   // Also there could be an empty byte at beginning from the toByteArray() call
   // as an inital zero may have been added to ensure the number is positive.

   if (fractionBytes.length > 14)
   {
      byte[] newBytes = new byte[14];
      int i = fractionBytes.length - 14;
      System.arraycopy (fractionBytes, i, newBytes, 0, 14);
      fractionBytes = newBytes;
   }

   int offset = 1;
   if (numShifts < 0)
   {
      byte[] newBytes = new byte[14];
      int i =  14 - fractionBytes.length;
      System.arraycopy (fractionBytes, 0, newBytes, i, fractionBytes.length);
      fractionBytes = newBytes;
   }

   // Bits 8 to 63 and 72 to 127 contain the fraction or mantissa (112 bits)
   System.arraycopy (fractionBytes, 0, result, 1, 7);
   System.arraycopy (fractionBytes, 7, result, 9, 7);

   // The bits 1 to 7 (7 bits) contain the biased exponent (exponent + bias)

   byte exponentByte = (byte)biasedExponent;
   result[0] = (byte)(exponentByte | result[0]);

   return (result);
}


/*
 * Long HFP Number
 *
 *   _ ______________ ________/____________
 *  | |              |                     |
 *  |S|Characteristic|   14-Digit Fraction |
 *  |_|______________|________/____________|
 *  0  1              8                   63
 *
 * The input byte array is expected to be in Big Endian format
 */

public static String convertIBM390HexDoubleToString (byte[] bytes)
{
   // The first bit is the sign bit

   boolean isPositive = true;
   if ((bytes[0] & BYTE_SIGN_MASK) != 0)
      isPositive = false;

   // The bits 1 to 7 (7 bits) contain the characteristic or biased exponent (exponent + bias)
   byte exponent = bytes[0];

   // Mask out the sign bit
   exponent = (byte)(exponent & (byte)0x7F);

   // Subtract 64 from the exponent to get the real exponent (range of 0 to 127)
   byte realExponent = (byte)(exponent - 64);

   // Bits 8 to 63 contain the fraction or mantissa (56 bits)
   // Allocate an extra byte so the BigInteger created will always be positive,
   // as we cannot right shift with a zero extension on a BigInteger.
   byte[] fractionBytes = new byte[8];
   System.arraycopy (bytes, 1, fractionBytes, 1, 7);

   BigInteger fraction = new BigInteger(fractionBytes);

   // Handle +- 0 and +- Infinity
   if (exponent != 0)
   {
      if (exponent == 0x7F)   // +- Infinity or NaN
      {
         if (!fraction.equals(BigInteger.ZERO)) // NaN
            return ("NaN");
         else if (isPositive)
            return ("Infinity");
         else
            return ("-Infinity");
      }
   }
   else  // Unnormalized number or +- 0 if (fraction == 0)
   {
      if (fraction.equals(BigInteger.ZERO))
      {
         if (isPositive)
            return ("0");
         else
            return ("-0");
      }
   }

   // Need to shift the fraction bits to the right up to the first bit set.
   // As the exponent is base 16 rather than base 2, we need to shift right
   // in groups of four bits (divide by 16), to reduce the exponent by one.
   int firstNonZero = fraction.getLowestSetBit();

   int numShift = 56;      // Number of fraction bits
   if (exponent == 0)      // Denormal number
   {
      // Do not need to shift the fraction bits as there is no implicit one, but
      // need to reduce the exponent by a value of the lowest bit set in a group of 4.
      // The first non zero set will always be on a 4 bit boundary.
      numShift = numShift - firstNonZero;       // Number of bits shifting number to the left
      while ((numShift % 4) != 0)
      {
         numShift--;
         fraction = fraction.shiftLeft (1);
      }
      realExponent -= numShift / 4;
   }
   else
   {
      if (firstNonZero != 0)
      {
         firstNonZero--;      // Number of bits need to shift fraction over
         while ((firstNonZero % 4) != 0)
            firstNonZero--;

         numShift = 56 - firstNonZero;    // Must shift in groups of 4 bits to reduce exponent by 1
         fraction = fraction.shiftRight (firstNonZero);
      }
      realExponent -= numShift / 4;    // Reduce the exponent by the number of shifts divided by 4
   }

   String result = getFractionAsString (isPositive, realExponent, fraction, 16, 17);

   return (result);
}

public static byte[] convertBigDecimalToIBM390HexDouble (BigDecimal fpNumber)
{
   byte[] result = new byte[8];

   // Handle the special case of zero immediately
   if (fpNumber.signum() == 0)
      return (result);

   // The first bit is the sign bit
   boolean isPositive = true;

   if (fpNumber.signum() == -1)
   {
      isPositive = false;
      result[0] = (byte)(result[0] | BYTE_SIGN_MASK); // Set the sign bit
   }

   fpNumber = fpNumber.abs();    // No longer need sign
   BigInteger[] base2 = getBase2FractionAndExponent (fpNumber, 56, false);
   BigInteger realFraction = base2[0];

   // The exponent bias is 64 as the real exponent range is 0 to 127.
   // As the exponent is base 16 rather than base 2, we need to shift right
   // in groups of four bits (divide by 16), to reduce the exponent by one.
   // If the exponent bias - abs(exponent) < 0, we have a denormal number.
   // Set the exponent to zero and shift the fraction over.

   short realExponent = base2[1].shortValue();

   // Need to divide the exponent by 4 to get the base 16 exponent.
   // If it does not divide evenly, we need to keep shifting the fraction
   // to the right by one and increasing the exponent by one.

   while ((realExponent % 4) != 0)
   {
      realFraction = realFraction.shiftRight (1);
      realExponent += 1;
   }

   short base16Exponent = (short)(realExponent / 4);

   short biasedExponent = 64;
   int numShifts = biasedExponent - Math.abs(base16Exponent);

   if (numShifts < 0)   // Denormal - fix fraction and exponent
   {
      realFraction = realFraction.shiftRight (Math.abs(numShifts*4));
      biasedExponent = 0;
   }
   else
      biasedExponent += base16Exponent;

   byte[] fractionBytes = realFraction.toByteArray();

   // If the number is denormal, we won't get back 7 bytes from the above call,
   // so we need to put them into the correct spot in the byte array.
   // Also there could be an empty byte at beginning from the toByteArray() call
   // as an inital zero may have been added to ensure the number is positive.

   if (fractionBytes.length > 7)
   {
      byte[] newBytes = new byte[7];
      int i = fractionBytes.length - 7;
      System.arraycopy (fractionBytes, i, newBytes, 0, 7);
      fractionBytes = newBytes;
   }

   int offset = 1;
   if (numShifts < 0)
      offset = 8 - fractionBytes.length;

   // Bits 8 to 63 contain the fraction or mantissa (56 bits)
   System.arraycopy (fractionBytes, 0, result, offset, Math.min (fractionBytes.length,7));

   // The bits 1 to 7 (7 bits) contain the biased exponent (exponent + bias)

   byte exponentByte = (byte)biasedExponent;
   result[0] = (byte)(exponentByte | result[0]);

   return (result);
}

}
